from datetime import timedelta

from django.core.management.base import BaseCommand
from django.utils import timezone

from security.enums import CeleryTaskRunLogState, CeleryTaskLogState
from security.models import CeleryTaskLog

from celery import current_app
from celery.exceptions import NotRegistered


class Command(BaseCommand):

    def handle(self, **options):
        for task in CeleryTaskLog.objects.filter_processing(stale__lt=timezone.now()).iterator():
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                task.change_and_save(
                    state=CeleryTaskLogState.SUCCEEDED
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                task.change_and_save(
                    state=CeleryTaskLogState.FAILED
                )
            else:
                try:
                    current_app.tasks[task.name].expire_task(task)
                except NotRegistered:
                    task.change_and_save(
                        state=CeleryTaskLogState.EXPIRED
                    )



from django.apps import AppConfig
from django.db.models.signals import pre_migrate, post_migrate

from django_celery_extensions.config import DEFAULTS

from . import utils

# Patch django_celery_extensions configuration
DEFAULTS['AUTO_GENERATE_TASKS_BASE'] = 'security.task.LoggedTask'

def start_migration(sender, **kwargs):
    utils.is_running_migration = True


def end_migration(sender, **kwargs):
    utils.is_running_migration = False


class SecurityLoggerAppConfig(AppConfig):

    name = 'security'
    verbose_name = 'Security'

    def ready(self):
        pre_migrate.connect(start_migration, sender=self)
        post_migrate.connect(end_migration, sender=self)



import re
import json
from json import JSONDecodeError

from django.conf import settings as django_settings
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.template.defaultfilters import truncatechars
from django.urls import resolve
from django.urls.exceptions import Resolver404
from django.utils import timezone
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import localtime

try:
    from django.core.urlresolvers import get_resolver
except ImportError:
    from django.urls import get_resolver

from generic_m2m_field.models import MultipleDBGenericManyToManyField

from chamber.models import SmartModel, SmartQuerySet

from enumfields import NumEnumField

from security.config import settings
from security.utils import get_headers, remove_nul_from_string, regex_sub_groups_global, is_base_collection

try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey

try:
    from pyston.filters.default_filters import CaseSensitiveStringFieldFilter
except ImportError:
    CaseSensitiveStringFieldFilter = object

from .compatibility import get_client_ip
from .enums import InputLoggedRequestType, LoggedRequestStatus, CeleryTaskLogState, CeleryTaskRunLogState


def display_json(value, indent=4):
    return json.dumps(value, indent=indent, ensure_ascii=False, cls=DjangoJSONEncoder)


def hide_sensitive_data_body(content):
    if settings.HIDE_SENSITIVE_DATA:
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('BODY', ()):
            content = regex_sub_groups_global(pattern, settings.SENSITIVE_DATA_REPLACEMENT, content)
    return content


def hide_sensitive_data_headers(headers):
    if settings.HIDE_SENSITIVE_DATA:
        headers = dict(headers)
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('HEADERS', ()):
            for header_name, header in headers.items():
                if re.match(pattern, header_name, re.IGNORECASE):
                    headers[header_name] = settings.SENSITIVE_DATA_REPLACEMENT
    return headers


def hide_sensitive_data_queries(queries):
    if settings.HIDE_SENSITIVE_DATA:
        queries = dict(queries)
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('QUERIES', ()):
            for query_name, query in queries.items():
                if re.match(pattern, query_name, re.IGNORECASE):
                    queries[query_name] = (
                        len(query) * [settings.SENSITIVE_DATA_REPLACEMENT] if is_base_collection(query)
                        else settings.SENSITIVE_DATA_REPLACEMENT
                    )
    return queries


def truncate_json_data(data):
    if isinstance(data, dict):
        return {key: truncate_json_data(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [truncate_json_data(val) for val in data]
    elif isinstance(data, str):
        return truncatechars(data, settings.LOG_JSON_STRING_LENGTH)
    else:
        return data


def truncate_body(content, max_length):
    content = force_text(content, errors='replace')
    if len(content) > max_length:
        try:
            json_content = json.loads(content)
            return (
                json.dumps(truncate_json_data(json_content))
                if isinstance(json_content, (dict, list)) and settings.LOG_JSON_STRING_LENGTH is not None
                else content[:max_length + 1]
            )
        except JSONDecodeError:
            return content[:max_length + 1]
    else:
        return content


def clean_body(body, max_length):
    cleaned_body = truncatechars(
        truncate_body(body, max_length), max_length + len(settings.SENSITIVE_DATA_REPLACEMENT)
    ) if max_length is not None else str(body)
    cleaned_body = hide_sensitive_data_body(remove_nul_from_string(cleaned_body)) if cleaned_body else cleaned_body
    cleaned_body = truncatechars(cleaned_body, max_length) if max_length else cleaned_body
    return cleaned_body


def clean_headers(headers):
    return hide_sensitive_data_headers(headers) if headers else headers


def clean_queries(queries):
    return hide_sensitive_data_queries(queries) if queries else queries


class BaseLogQuerySet(SmartQuerySet):

    def filter_related_with_object(self, related_object):
        return self.filter(
            related_objects__object_id=related_object.pk,
            related_objects__object_ct_id=ContentType.objects.get_for_model(related_object).pk
        )


class InputLoggedRequestManager(models.Manager):
    """
    Create new LoggedRequest instance from HTTP request
    """

    def prepare_from_request(self, request):
        user_pk = request.user.pk if hasattr(request, 'user') and request.user.is_authenticated else None
        path = truncatechars(request.path, 200)

        try:
            slug = resolve(request.path_info, getattr(request, 'urlconf', None)).view_name
        except Resolver404:
            slug = None

        return self.model(
            request_headers=clean_headers(get_headers(request)),
            request_body=clean_body(request.body, settings.LOG_REQUEST_BODY_LENGTH),
            user_id=user_pk,
            method=request.method.upper()[:7],
            host=request.get_host(),
            path=path,
            queries=clean_queries(request.GET.dict()),
            is_secure=request.is_secure(),
            ip=get_client_ip(request)[0],
            request_timestamp=timezone.now(),
            slug=slug
        )


class LoggedRequest(SmartModel):

    host = models.CharField(_('host'), max_length=255, null=False, blank=False, db_index=True)
    host._filter = CaseSensitiveStringFieldFilter
    method = models.SlugField(_('method'), max_length=7, null=False, blank=False, db_index=True)
    path = models.CharField(_('URL path'), max_length=2000, null=False, blank=True, db_index=True)
    path._filter = CaseSensitiveStringFieldFilter
    queries = models.JSONField(_('queries'), null=True, blank=True, encoder=DjangoJSONEncoder)
    is_secure = models.BooleanField(_('HTTPS connection'), default=False, null=False, blank=False)
    slug = models.CharField(_('slug'), null=True, blank=True, db_index=True, max_length=255)

    # Request information
    request_timestamp = models.DateTimeField(_('request timestamp'), null=False, blank=False, db_index=True)
    request_headers = models.JSONField(_('request headers'), null=True, blank=True, encoder=DjangoJSONEncoder)
    request_body = models.TextField(_('request body'), null=False, blank=True)

    # Response information
    response_timestamp = models.DateTimeField(_('response timestamp'), null=True, blank=True, db_index=True)
    response_code = models.PositiveSmallIntegerField(_('response code'), null=True, blank=True)
    response_headers = models.JSONField(_('response headers'), null=True, blank=True, encoder=DjangoJSONEncoder)
    response_body = models.TextField(_('response body'), null=True, blank=True)
    response_time = models.FloatField(_('response time'), null=True, blank=True)

    status = NumEnumField(verbose_name=_('status'), enum=LoggedRequestStatus, null=False, blank=False,
                          default=LoggedRequestStatus.INCOMPLETE)
    error_description = models.TextField(_('error description'), null=True, blank=True)
    exception_name = models.CharField(_('exception name'), null=True, blank=True, max_length=255)

    @classmethod
    def get_status(cls, status_code):
        if status_code >= 500:
            return LoggedRequestStatus.ERROR
        elif status_code >= 400:
            return LoggedRequestStatus.WARNING
        else:
            return LoggedRequestStatus.INFO

    def short_queries(self):
        return truncatechars(display_json(self.queries, indent=0), 50)
    short_queries.short_description = _('queries')
    short_queries.filter_by = 'queries'

    def short_request_headers(self):
        return truncatechars(display_json(self.request_headers, indent=0), 50)
    short_request_headers.short_description = _('request headers')
    short_request_headers.filter_by = 'request_headers'

    def short_path(self):
        return truncatechars(self.path, 50)
    short_path.short_description = _('Path')
    short_path.filter_by = 'path'
    short_path.order_by = 'path'

    def short_response_body(self):
        return truncatechars(self.response_body, 50) if self.response_body is not None else None
    short_response_body.short_description = _('response body')
    short_response_body.filter_by = 'response_body'

    def short_request_body(self):
        return truncatechars(self.request_body, 50)
    short_request_body.short_description = _('request body')
    short_request_body.filter_by = 'request_body'

    def __str__(self):
        return ' '.join(
            (force_text(v) for v in (
                self.slug, self.response_code, localtime(self.request_timestamp.replace(microsecond=0))
            ) if v)
        )

    class Meta:
        abstract = True

    class UIMeta:
        default_ui_filter_by = 'id'


class InputLoggedRequest(LoggedRequest):

    user_id = models.TextField(verbose_name=_('user ID'), null=True, blank=True, db_index=True)
    ip = models.GenericIPAddressField(_('IP address'), null=False, blank=False, db_index=True)
    type = NumEnumField(verbose_name=_('type'), enum=InputLoggedRequestType,
                        default=InputLoggedRequestType.COMMON_REQUEST,
                        null=False, blank=False, db_index=True)
    related_objects = MultipleDBGenericManyToManyField()

    objects = InputLoggedRequestManager.from_queryset(BaseLogQuerySet)()

    class Meta:
        verbose_name = _('input logged request')
        verbose_name_plural = _('input logged requests')
        ordering = ('-created_at',)

    class SmartMeta:
        is_cleaned_pre_save = False

    @cached_property
    def user(self):
        return get_user_model().objects.filter(pk=self.user_id).first()

    def update_from_response(self, response):
        self.response_timestamp = timezone.now()
        self.response_time = (self.response_timestamp - self.request_timestamp).total_seconds()
        self.status = self.get_status(response.status_code)
        self.response_code = response.status_code
        self.response_headers = clean_headers(dict(response.items()))

        if (not response.streaming and settings.LOG_RESPONSE_BODY_CONTENT_TYPES is not None and
                response.get('content-type', '').split(';')[0] in settings.LOG_RESPONSE_BODY_CONTENT_TYPES):
            response_body = clean_body(response.content, settings.LOG_RESPONSE_BODY_LENGTH)
        else:
            response_body = ''

        self.response_body = response_body


class OutputLoggedRequest(LoggedRequest):

    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        verbose_name = _('output logged request')
        verbose_name_plural = _('output logged requests')
        ordering = ('-created_at',)

    class SmartMeta:
        is_cleaned_pre_save = False


class CommandLog(SmartModel):
    """
    Represents a log of a command run.

    Attributes:
        start: Date and time when command was started.
        stop: Date and time when command finished.
        time: Command processing time in miliseconds.
        name: Name of the command.
        options: Arguments/options the command was run with.
        executed_from_command_line: Flag that indicates if command was run from the command line.
        output: Standard and error output of the command.
        is_successful: Flag that indicates if command finished successfully.
    """
    start = models.DateTimeField(verbose_name=_('start'), blank=False, null=False, editable=False)
    stop = models.DateTimeField(verbose_name=_('stop'), blank=True, null=True, editable=False)
    time = models.FloatField(verbose_name=_('time'), null=True, blank=True)
    name = models.CharField(max_length=250, blank=False, null=False, editable=False, db_index=True,
                            verbose_name=_('name'))
    input = models.TextField(verbose_name=_('input'), blank=False, null=False, editable=False)
    executed_from_command_line = models.BooleanField(verbose_name=_('executed from command line'),
                                                     blank=False, null=False, default=False, editable=False)
    output = models.TextField(verbose_name=_('output'), blank=True, null=True, editable=False)
    is_successful = models.BooleanField(verbose_name=_('finished successfully'),
                                        blank=False, null=False, default=False, editable=False)
    error_message = models.TextField(verbose_name=_('error message'), null=True, blank=True, editable=False)
    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        verbose_name = _('command log')
        verbose_name_plural = _('command logs')
        ordering = ('-created_at',)

    class UIMeta:
        default_ui_filter_by = 'id'


class CeleryTaskLogManager(models.Manager):

    def filter_waiting(self):
        return self.filter(
            state=CeleryTaskLogState.WAITING
        )

    def filter_processing(self, related_objects=None, **kwargs):
        qs = self.filter(
            state__in={
                CeleryTaskLogState.ACTIVE,
                CeleryTaskLogState.WAITING,
                CeleryTaskLogState.RETRIED
            }
        ).filter(**kwargs)

        if related_objects:
            for related_object in related_objects:
                qs = qs.filter(
                    related_objects__object_id=related_object.pk,
                    related_objects__object_ct_id=ContentType.objects.get_for_model(related_object).pk
                )
        return qs


class CeleryTaskLog(SmartModel):

    celery_task_id = models.CharField(
        verbose_name=_('celery ID'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    name = models.CharField(
        verbose_name=_('task name'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    queue_name = models.CharField(
        verbose_name=_('queue name'),
        null=True,
        blank=True,
        max_length=250
    )
    input = models.TextField(
        verbose_name=_('input'),
        blank=True,
        null=True,
        editable=False
    )
    task_args = models.JSONField(
        verbose_name=_('task args'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        verbose_name=_('task kwargs'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    estimated_time_of_first_arrival = models.DateTimeField(
        verbose_name=_('estimated time of first arrival'),
        null=False,
        blank=False
    )
    expires = models.DateTimeField(
        verbose_name=_('time of expiration'),
        null=True,
        blank=True
    )
    stale = models.DateTimeField(
        verbose_name=_('stale task time'),
        null=True,
        blank=True
    )
    state = NumEnumField(
        verbose_name=_('state'),
        null=False,
        blank=False,
        enum=CeleryTaskLogState,
        default=CeleryTaskLogState.WAITING,
        db_index=True
    )

    related_objects = MultipleDBGenericManyToManyField()

    objects = CeleryTaskLogManager.from_queryset(BaseLogQuerySet)()

    class Meta:
        verbose_name = _('celery task')
        verbose_name_plural = _('celery tasks')
        ordering = ('-created_at',)

    class UIMeta:
        default_ui_filter_by = 'id'

    def __str__(self):
        return '{} ({})'.format(self.name, self.state.label)

    @property
    def runs(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id)

    @property
    def last_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).last()

    @property
    def first_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).first()

    def get_start(self):
        first_run = self.first_run
        return first_run.start if first_run else None
    get_start.short_description = _('start')

    def get_stop(self):
        last_run = self.last_run
        return last_run.stop if last_run else None
    get_stop.short_description = _('stop')


class CeleryTaskRunLog(SmartModel):

    celery_task_id = models.CharField(
        verbose_name=_('celery ID'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    start = models.DateTimeField(
        verbose_name=_('start'),
        null=True,
        blank=True
    )
    stop = models.DateTimeField(
        verbose_name=_('stop'),
        null=True,
        blank=True
    )
    time = models.FloatField(
        verbose_name=_('time'),
        null=True,
        blank=True
    )
    state = NumEnumField(
        verbose_name=_('state'),
        null=False,
        blank=False,
        enum=CeleryTaskRunLogState,
        default=CeleryTaskRunLogState.ACTIVE,
        db_index=True
    )
    name = models.CharField(
        verbose_name=_('task name'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    task_args = models.JSONField(
        verbose_name=_('task args'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        verbose_name=_('task kwargs'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    result = models.JSONField(
        verbose_name=_('result'),
        blank=True,
        null=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    error_message = models.TextField(
        verbose_name=_('error message'),
        null=True,
        blank=True
    )
    output = models.TextField(
        verbose_name=_('output'),
        blank=True,
        null=True,
        editable=False
    )
    retries = models.PositiveSmallIntegerField(
        verbose_name=_('retries'),
        null=False,
        blank=False,
        default=0
    )
    estimated_time_of_next_retry = models.DateTimeField(
        verbose_name=_('estimated time of arrival'),
        null=True,
        blank=True
    )
    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        verbose_name = _('celery task run')
        verbose_name_plural = _('celery tasks run')
        ordering = ('created_at',)

    def __str__(self):
        return '{} ({})'.format(self.name, self.get_state_display())

    def get_task_log(self):
        return CeleryTaskLog.objects.filter(celery_task_id=self.celery_task_id).first()
    get_task_log.short_description = _('celery task')



import os
import json
import gzip
import math

from datetime import datetime, time, timedelta

from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.core.management.base import CommandError
from django.core.management.base import BaseCommand
from django.utils.timezone import now, utc
from django.utils.module_loading import import_string

from security.config import settings
from security.models import InputLoggedRequest, OutputLoggedRequest, CommandLog, CeleryTaskLog, CeleryTaskRunLog

from io import TextIOWrapper


storage = import_string(settings.BACKUP_STORAGE_CLASS)()


UNIT_OPTIONS = {
    'h': lambda amount: now() - timedelta(hours=amount),
    'd': lambda amount: now() - timedelta(days=amount),
    'w': lambda amount: now() - timedelta(weeks=amount),
    'm': lambda amount: now() - timedelta(days=(30 * amount)),  # 30-day month
    'y': lambda amount: now() - timedelta(weeks=(52 * amount)),  # 364-day year
}


def lazy_serialize_qs_without_pk(qs):
    def generator():
        for obj in qs.iterator():
            data = serializers.serialize('python', [obj])[0]

            del data['pk']
            yield data

    class StreamList(list):
        def __iter__(self):
            return generator()

        # according to the comment below
        def __len__(self):
            return 1

    return StreamList()


def get_querysets_by_batch(qs, batch):
    steps = math.ceil(qs.count() / batch)
    for _ in range(steps):
        yield qs[:batch]


class Command(BaseCommand):

    models = {
        'input-request': InputLoggedRequest,
        'output-request': OutputLoggedRequest,
        'command': CommandLog,
        'celery': CeleryTaskLog,
        'celery-run': CeleryTaskRunLog,
    }

    def add_arguments(self, parser):
        parser.add_argument('--type', action='store', dest='type', default='input-request',
                            help='Tells Django what type of requests should be removed '
                                 '({}).'.format('/'.join(self.models.keys())))
        parser.add_argument('--expiration', action='store', dest='expiration',
                            help='Sets the timedelta from which logs will be removed.', required=True)
        parser.add_argument('--noinput', action='store_false', dest='interactive', default=True,
                            help='Tells Django to NOT prompt the user for input of any kind.')
        parser.add_argument('--backup', action='store', dest='backup', default=False,
                            help='Tells Django where to backup removed logs.')

    def backup_logs_and_clean_data(self, qs, backup_path):
        for timestamp in qs.datetimes('created_at', 'day', tzinfo=utc):
            min_timestamp = datetime.combine(timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(timestamp, time.max).replace(tzinfo=utc)
            qs_filtered_by_day = qs.filter(created_at__range=(min_timestamp, max_timestamp))

            for qs_batch in get_querysets_by_batch(qs_filtered_by_day, settings.PURGE_LOG_BACKUP_BATCH):
                self.stdout.write(
                    2 * ' ' + 'Cleaning logs for date {} ({})'.format(
                        timestamp.date(), qs_batch.count()
                    )
                )
                if backup_path:
                    log_file_name = os.path.join(backup_path, str(timestamp.date()))

                    if storage.exists('{}.json.gz'.format(log_file_name)):
                        i = 1
                        while storage.exists('{}({}).json.gz'.format(log_file_name, i)):
                            i += 1
                        log_file_name = '{}({})'.format(log_file_name, i)

                    self.stdout.write(4 * ' ' + 'generating backup file: {}.json.gz'.format(log_file_name))

                    with storage.open('{}.json.gz'.format(log_file_name), 'wb') as f:
                        with TextIOWrapper(gzip.GzipFile(filename='{}.json'.format(log_file_name),
                                                         fileobj=f, mode='wb')) as gzf:
                            json.dump(
                                lazy_serialize_qs_without_pk(qs_batch), gzf, cls=DjangoJSONEncoder, indent=5
                            )
                self.stdout.write(4 * ' ' + 'deleting logs')

                for qs_batch_to_delete in get_querysets_by_batch(qs_batch, settings.PURGE_LOG_DELETE_BATCH):
                    qs_filtered_by_day.filter(pk__in=qs_batch_to_delete).delete()

    def handle(self, expiration, type, **options):
        # Check we have the correct values
        unit = expiration[-1]
        amount = expiration[0:-1]
        try:
            amount = int(amount)
        except ValueError:
            raise CommandError('Invalid expiration format')

        if unit not in UNIT_OPTIONS:
            raise CommandError('Invalid expiration format')

        model = self.get_model(type)
        qs = model.objects.filter(created_at__lte=UNIT_OPTIONS[unit](amount)).order_by('created_at', 'pk')

        if qs.count() == 0:
            self.stdout.write('There are no logs to delete.')
        else:
            if options.get('interactive'):
                confirm = input('''
                   You have requested a database reset.
                   This will IRREVERSIBLY DESTROY any
                   logs created before {} {}
                   ago. That is a total of {} logs.
                   Are you sure you want to do this?
                   Type 'yes' to continue, or 'no' to cancel: '''.format(amount, unit, qs.count()))
            else:
                confirm = 'yes'

            if confirm == 'yes':
                try:
                    self.stdout.write('Clean data')
                    self.backup_logs_and_clean_data(qs, options.get('backup'))
                except IOError as ex:
                    self.stderr.write(str(ex))

    def get_model(self, type):
        model = self.models.get(type)
        if model:
            return model
        else:
            raise CommandError('Type can be only one of "{}"'.format('/'.join(self.models.keys())))



from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import FileSystemStorage
from django.utils.functional import cached_property

from security.config import settings


# File system storage
class BackupFileSystemStorage(FileSystemStorage):

    @cached_property
    def base_location(self):
        if not settings.BACKUP_STORAGE_PATH:
            raise ImproperlyConfigured('SECURITY_BACKUP_STORAGE_PATH settings is not set')
        return settings.BACKUP_STORAGE_PATH



import io
import json
import sys
import uuid
from distutils.version import StrictVersion
from unittest import mock

import responses
from celery.exceptions import CeleryError, TimeoutError
from freezegun import freeze_time

from django import get_version
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import CommandError
from django.db import transaction
from django.test import override_settings
from django.utils.timezone import now, timedelta

from germanium.decorators import data_consumer, data_provider
from germanium.test_cases.client import ClientTestCase
from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import (
    assert_equal, assert_false, assert_http_not_found, assert_http_ok, assert_http_redirect,
    assert_http_too_many_requests, assert_in, assert_is_none, assert_is_not_none, assert_not_in, assert_not_raises,
    assert_raises, assert_true
)

from django_celery_extensions.task import get_django_command_task

from security.config import settings
from security.decorators import atomic_log
from security.management import call_command
from security.models import (
    CeleryTaskLog, CeleryTaskLogState, CeleryTaskRunLog, CeleryTaskRunLogState, CommandLog, InputLoggedRequest,
    LoggedRequestStatus, OutputLoggedRequest
)
from security.transport import security_requests as requests
from security.utils import log_context_manager

from apps.test_security.tasks import error_task, retry_task, sum_task, unique_task


TRUNCATION_CHAR = '…' if StrictVersion(get_version()) > StrictVersion('2.2') else '...'
TRUNCATION_DIFF = 0 if StrictVersion(get_version()) > StrictVersion('2.2') else -2


class TestException(Exception):
    pass


class BaseTestCaseMixin:

    databases = ['default', 'security']

    @data_provider
    def create_user(self, username='test', email='test@test.cz'):
        return User.objects._create_user(username, email, 'test', is_staff=True, is_superuser=True)


class CeleryHealthCheckCommandTestCaseMixin(BaseTestCaseMixin):

    @data_provider
    def create_waiting_celery_task_log(self, task_uuid=uuid.uuid4(), name=None,
                                       estimated_time_of_first_arrival=now() + timedelta(days=1),
                                       queue_name=settings.CELERY_HEALTH_CHECK_DEFAULT_QUEUE,
                                       **kwargs):
        name = 'test_task' if name is None else name
        return CeleryTaskLog.objects.create(
            celery_task_id=task_uuid,
            name=name,
            estimated_time_of_first_arrival=estimated_time_of_first_arrival,
            queue_name=queue_name,
            **kwargs
        )


def test_call_command(*args, **kwargs):
    call_command(*args, **kwargs, stdout=io.StringIO(), stderr=io.StringIO())


class SecurityTestCase(BaseTestCaseMixin, ClientTestCase):

    @mock.patch('security.transport.security_requests.log_output_request')
    def test_every_output_request_has_data_for_stdout_logging(self, func):
        requests.get('http://test.cz')

        assert_true(func.called)
        func_args = func.call_args.args[0] if sys.version_info >= (3, 8) else func.call_args_list[0][0][0]  # data
        assert_in('request_timestamp', func_args)
        assert_in('response_timestamp', func_args)
        assert_in('response_time', func_args)
        assert_in('response_code', func_args)
        assert_in('host', func_args)
        assert_in('path', func_args)
        assert_in('method', func_args)
        assert_in('slug', func_args)

    def test_every_request_should_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 1)

    @data_consumer('create_user')
    def test_data_change_should_be_connected_with_logged_request(self, user):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        assert_equal(InputLoggedRequest.objects.count(), 1)
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(
            input_logged_request.related_objects.get().object.version_set.get().content_type,
            ContentType.objects.get_for_model(User)
        )

    def test_data_change_should_be_connected_with_celery_task_run_log(self):
        get_django_command_task('create_user').apply_async()
        assert_equal(CeleryTaskRunLog.objects.count(), 1)
        celery_task_run_log = CeleryTaskRunLog.objects.get()
        assert_equal(
            celery_task_run_log.related_objects.get().object.version_set.get().content_type,
            ContentType.objects.get_for_model(User)
        )

    def test_data_change_should_be_connected_with_command_log(self):
        test_call_command('create_user')
        assert_equal(CommandLog.objects.count(), 1)
        command_log = CommandLog.objects.get()
        assert_equal(
            command_log.related_objects.get().object.version_set.get().content_type,
            ContentType.objects.get_for_model(User)
        )

    @data_consumer('create_user')
    def test_input_logged_request_should_have_right_status(self, user):
        assert_http_ok(self.post('/admin/login/', data={'username': 'invalid', 'password': 'invalid'}))
        assert_equal(InputLoggedRequest.objects.first().status, LoggedRequestStatus.INFO)
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        assert_equal(InputLoggedRequest.objects.first().status, LoggedRequestStatus.INFO)
        assert_raises(Exception, self.get, '/proxy/')
        assert_equal(InputLoggedRequest.objects.first().status, LoggedRequestStatus.ERROR)
        assert_http_not_found(self.get('/404/'))
        assert_equal(InputLoggedRequest.objects.first().status, LoggedRequestStatus.WARNING)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_IP=('127.0.0.1',))
    def test_ignored_client_ip_should_not_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_URL_PATHS=('/',))
    def test_ignored_request_path_should_not_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        self.get('/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=10)
    def test_request_body_should_be_truncated(self):
        self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 10)
        assert_true(input_logged_request.request_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=10)
    def test_response_body_should_be_truncated(self):
        self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.response_body), 10)
        assert_true(input_logged_request.response_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=None)
    def test_request_body_truncation_should_be_turned_off(self):
        self.post('/admin/login/', data={'username': 2000 * 'a', 'password': 2000 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 4183)
        assert_false(input_logged_request.request_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=None)
    def test_response_body_truncation_should_be_turned_off(self):
        resp = self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.response_body, str(resp.content))
        assert_false(input_logged_request.response_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=())
    def test_not_allowed_content_type_should_not_be_logged(self):
        self.get('/')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.response_body)

    @override_settings(SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=None)
    def test_allowed_content_type_should_be_turned_off(self):
        self.get('/')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.response_body)

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_should_be_truncated_with_another_method(self):
        self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a', 'b': 50 * 'b'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(
            json.loads(input_logged_request.request_body),
            json.loads('{"a": "%s%s", "b": "%s%s"}' % ((9 + TRUNCATION_DIFF) * 'a', TRUNCATION_CHAR, (9 + TRUNCATION_DIFF) * 'b', TRUNCATION_CHAR))
        )
        assert_false(input_logged_request.request_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=50, SECURITY_LOG_JSON_STRING_LENGTH=None)
    def test_json_request_should_not_be_truncated_with_another_method(self):
        self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()

        assert_equal(input_logged_request.request_body, '{"a": "' + (42 + TRUNCATION_DIFF) * 'a' + TRUNCATION_CHAR)
        assert_true(input_logged_request.request_body.endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_should_be_truncated_with_another_method_and_standard_method_too(self):
        self.c.post('/admin/login/', data=json.dumps({50 * 'a': 50 * 'a', 50 * 'b': 50 * 'b'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(len(input_logged_request.request_body), 100)
        assert_true(input_logged_request.request_body.endswith(TRUNCATION_CHAR))

    def test_response_with_exception_should_be_logged(self):
        assert_equal(InputLoggedRequest.objects.count(), 0)
        assert_raises(Exception, self.get, '/proxy/')
        assert_equal(InputLoggedRequest.objects.count(), 1)

    @responses.activate
    def test_response_without_exception_should_be_logged(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        assert_equal(self.get('/proxy/?url=http://test.cz').content, b'test')
        assert_equal(InputLoggedRequest.objects.count(), 1)
        assert_equal(OutputLoggedRequest.objects.count(), 1)

    @responses.activate
    @data_consumer('create_user')
    def test_output_logged_request_should_be_related_with_object(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        responses.add(responses.GET, 'http://test.cz', body='test')
        assert_equal(self.get('/proxy/?url=http://test.cz').content, b'test')
        assert_equal(InputLoggedRequest.objects.count(), 2)
        assert_equal(OutputLoggedRequest.objects.count(), 1)
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.related_objects.count(), 2)
        user_related_object, input_request_related_object = output_logged_request.related_objects.all()
        assert_equal(user_related_object.object, user)
        assert_equal(input_request_related_object.object, InputLoggedRequest.objects.first())

    def test_sensitive_data_body_in_json_should_be_hidden(self):
        self.c.post('/admin/login/', data=json.dumps({'username': 'test', 'password': 'secret-password'}),
                    content_type='application/json')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_in('"password": "[Filtered]"', input_logged_request.request_body)
        assert_not_in('"password": "secret-password"', input_logged_request.request_body)

    def test_sensitive_data_body_in_raw_form_should_be_hidden(self):
        self.post('/admin/login/', data={'username': 'test', 'password': 'secret-password\nddd'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_in('[Filtered]', input_logged_request.request_body)

    @data_consumer('create_user')
    def test_sensitive_headers_should_be_hidden(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.request_headers['COOKIE'], '[Filtered]')

    def test_sensitive_queries_should_be_hidden(self):
        self.get('/?token=test')
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.queries['token'], '[Filtered]')

    @data_consumer('create_user')
    @override_settings(SECURITY_SENSITIVE_DATA_REPLACEMENT='(Filtered)')
    def test_sensitive_replacement_should_be_changed(self, user):
        assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
        input_logged_request = InputLoggedRequest.objects.get()
        assert_equal(input_logged_request.request_headers['COOKIE'], '(Filtered)')

    @override_settings(SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS=())
    def test_command_should_be_logged(self):
        assert_equal(CommandLog.objects.count(), 0)
        test_call_command('showmigrations')
        assert_equal(CommandLog.objects.count(), 1)

    @override_settings(SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS=('showmigrations',))
    def test_excluded_command_should_not_be_logged(self):
        assert_equal(CommandLog.objects.count(), 0)
        test_call_command('showmigrations')
        assert_equal(CommandLog.objects.count(), 0)

    def test_throttling_should_be_raised(self):
        for _ in range(20):
            assert_http_redirect(self.get('/admin/'))
        assert_http_too_many_requests(self.get('/admin/'))

    @override_settings(SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators')
    def test_throttling_configuration_should_be_changed_via_settings(self):
        for _ in range(2):
            assert_http_redirect(self.get('/admin/'))
        assert_http_too_many_requests(self.get('/admin/'))

    def test_decorated_view_with_hide_request_body_should_not_log_request_body(self):
        self.post('/hide-request-body/', data={'a': 20 * 'a', 'b': 20 * 'b'})
        input_logged_request = InputLoggedRequest.objects.get()
        assert_false(input_logged_request.request_body)

    def test_decorated_view_with_log_exempt_should_not_log_request(self):
        self.get('/log-exempt/')
        assert_equal(InputLoggedRequest.objects.count(), 0)

    def test_decorated_view_with_throttling_exempt_should_not_raise_throttling_exception(self):
        for _ in range(20):
            assert_http_ok(self.get('/throttling-exempt/'))
        assert_http_ok(self.get('/throttling-exempt/'))

    def test_decorated_view_with_throttling_should_raise_throttling_exception(self):
        assert_http_ok(self.get('/extra-throttling/'))
        assert_http_too_many_requests(self.get('/extra-throttling/'))

    @responses.activate
    def test_output_logged_requests_with_atomic_block_should_be_logged_if_exception_is_raised(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with transaction.atomic():
                requests.get('http://test.cz')
                assert_equal(OutputLoggedRequest.objects.count(), 1)
                raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 1)

    @responses.activate
    def test_output_logged_requests_with_atomic_and_log_atomic_block_should_be_logged_if_exception_is_raised(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with atomic_log():
                with transaction.atomic():
                    requests.get('http://test.cz')
                    assert_equal(OutputLoggedRequest.objects.count(), 0)
                    raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 1)

    @responses.activate
    def test_output_logged_requests_with_atomic_and_log_atomic_block_should_be_nested(self):
        responses.add(responses.GET, 'http://test.cz', body='test')
        with assert_raises(TestException):
            with atomic_log():
                with transaction.atomic():
                    requests.get('http://test.cz')
                    with atomic_log():
                        with transaction.atomic():
                            requests.get('http://test.cz')
                    assert_equal(OutputLoggedRequest.objects.count(), 0)
                    raise TestException
        assert_equal(OutputLoggedRequest.objects.count(), 2)

    @responses.activate
    def test_response_sensitive_data_body_in_json_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', data=json.dumps({'password': 'secret-password'}))
        output_logged_requst = OutputLoggedRequest.objects.get()
        assert_in('"password": "[Filtered]"', output_logged_requst.request_body)
        assert_not_in('"password": "secret-password"', output_logged_requst.request_body)
        assert_in('"password": "secret-password"', responses.calls[0].request.body)
        assert_not_in('"password": "[Filtered]"', responses.calls[0].request.body)

    @responses.activate
    def test_response_sensitive_headers_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', headers={'token': 'secret'})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.request_headers['token'], '[Filtered]')
        assert_equal(responses.calls[0].request.headers['token'], 'secret')

    @responses.activate
    def test_response_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', params={'token': 'secret'})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries['token'], '[Filtered]')
        assert_equal(responses.calls[0].request.url, 'http://test.cz/?token=secret')

    @responses.activate
    def test_response_more_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz', params={'token': ['secret', 'secret2']})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries['token'], ['[Filtered]', '[Filtered]'])
        assert_equal(responses.calls[0].request.url, 'http://test.cz/?token=secret&token=secret2')

    @responses.activate
    def test_response_sensitive_params_and_url_query_together_data_should_be_logged(self):
        responses.add(responses.POST, 'http://test.cz', body='test')
        requests.post('http://test.cz?a=1&a=2', params={'b': '6', 'a': '3', 'c': ['5']})
        output_logged_request = OutputLoggedRequest.objects.get()
        assert_equal(output_logged_request.queries, {'b': '6', 'a': ['3', '1', '2'], 'c': '5'})

    @data_consumer('create_user')
    def test_apply_task_with_related_objects_should_be_stored(self, user):
        sum_task.apply_async(args=(5, 8), related_objects=[user])
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.SUCCEEDED)
        assert_equal(CeleryTaskLog.objects.get().related_objects.get().object, user)

    def test_celery_task_should_be_logged(self):
        sum_task.apply_async(args=(5, 8))
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.SUCCEEDED)
        assert_equal(CeleryTaskRunLog.objects.count(), 1)
        assert_equal(CeleryTaskRunLog.objects.get().state, CeleryTaskRunLogState.SUCCEEDED)

    def test_celery_task_should_be_able_to_run_with_apply_async_on_commit(self):
        sum_task.apply_async_on_commit(args=(5, 8))
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.SUCCEEDED)
        assert_equal(CeleryTaskRunLog.objects.count(), 1)
        assert_equal(CeleryTaskRunLog.objects.get().state, CeleryTaskRunLogState.SUCCEEDED)

    def test_celery_error_task_should_be_set_as_failed_in_the_log(self):
        error_task.apply_async_on_commit()
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.FAILED)
        assert_equal(CeleryTaskRunLog.objects.count(), 1)
        assert_equal(CeleryTaskRunLog.objects.get().state, CeleryTaskRunLogState.FAILED)
        assert_is_not_none(CeleryTaskRunLog.objects.get().error_message)

    def test_django_command_should_be_run_via_task(self):
        get_django_command_task('check').apply_async()
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.SUCCEEDED)
        assert_equal(CeleryTaskRunLog.objects.count(), 1)
        assert_equal(CeleryTaskRunLog.objects.get().state, CeleryTaskRunLogState.SUCCEEDED)

    def test_retry_command_should_be_automatically_retried(self):
        retry_task.apply_async()
        assert_equal(CeleryTaskLog.objects.count(), 1)
        assert_equal(CeleryTaskLog.objects.get().state, CeleryTaskLogState.SUCCEEDED)
        assert_equal(CeleryTaskRunLog.objects.count(), 6)
        assert_equal(CeleryTaskRunLog.objects.values('celery_task_id').distinct().count(), 1)
        assert_equal(
            tuple(CeleryTaskRunLog.objects.order_by('created_at').values_list('state', flat=True)),
            (CeleryTaskRunLogState.RETRIED, CeleryTaskRunLogState.RETRIED, CeleryTaskRunLogState.RETRIED,
             CeleryTaskRunLogState.RETRIED, CeleryTaskRunLogState.RETRIED, CeleryTaskRunLogState.SUCCEEDED)
        )

    @freeze_time(now())
    def test_retry_command_should_be_delayed(self):
        retry_task.apply_async()
        assert_equal(CeleryTaskRunLog.objects.count(), 6)
        assert_equal(CeleryTaskLog.objects.first().estimated_time_of_first_arrival, now())
        assert_equal(
            tuple(CeleryTaskRunLog.objects.order_by('created_at').values_list('estimated_time_of_next_retry', flat=True)),
            (now() + timedelta(minutes=1), now() + timedelta(minutes=5),
             now() + timedelta(minutes=10), now() + timedelta(minutes=30), now() + timedelta(minutes=60), None)
        )

    @freeze_time(now())
    def test_retry_command_should_have_set_right_retry_value(self):
        retry_task.apply_async()
        assert_equal(CeleryTaskRunLog.objects.count(), 6)
        assert_equal(
            tuple(CeleryTaskRunLog.objects.order_by('created_at').values_list('retries', flat=True)),
            (0, 1, 2, 3, 4, 5)
        )

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30)
    def test_celery_task_should_have_rightly_set_stale_time(self):
        sum_task.apply_async(args=(5, 8))
        celery_task_log = CeleryTaskLog.objects.get()
        assert_equal(celery_task_log.estimated_time_of_first_arrival, now())
        assert_is_none(celery_task_log.expires)
        assert_equal(celery_task_log.stale, now() + timedelta(seconds=30))

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_celery_task_should_have_rightly_set_expires_time(self):
        sum_task.apply_async(args=(5, 8))
        celery_task_log = CeleryTaskLog.objects.get()
        assert_equal(celery_task_log.estimated_time_of_first_arrival, now())
        assert_equal(celery_task_log.expires, now() + timedelta(seconds=20))
        assert_equal(celery_task_log.stale, now() + timedelta(seconds=30))

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_celery_task_should_have_rightly_set_expires_time_if_soft_time_limit_is_set_in_task_call(self):
        sum_task.apply_async(args=(5, 8), time_limit=20)
        celery_task_log = CeleryTaskLog.objects.get()
        assert_equal(celery_task_log.estimated_time_of_first_arrival, now())
        assert_equal(celery_task_log.expires, now() + timedelta(seconds=10))
        assert_equal(celery_task_log.stale, now() + timedelta(seconds=30))

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_celery_task_should_have_rightly_set_expires_time_if_stale_time_limit_is_set_in_task_call(self):
        sum_task.apply_async(args=(5, 8), stale_time_limit=100)
        celery_task_log = CeleryTaskLog.objects.get()
        assert_equal(celery_task_log.estimated_time_of_first_arrival, now())
        assert_equal(celery_task_log.stale, now() + timedelta(seconds=100))
        assert_equal(celery_task_log.expires, now() + timedelta(seconds=90))

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_celery_task_should_have_rightly_set_expires_time_according_to_default_task_stale_limit_value(self):
        error_task.apply_async()
        celery_task_log = CeleryTaskLog.objects.get()
        assert_equal(celery_task_log.estimated_time_of_first_arrival, now())
        assert_equal(celery_task_log.stale, now() + timedelta(seconds=60*60))
        assert_equal(celery_task_log.expires, now() + timedelta(seconds=60 * 60 - 10))

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_stale_waiting_celery_task_should_be_set_as_failed_with_command(self):
        sum_task.apply_async(args=(5, 8))
        celery_task_log = CeleryTaskLog.objects.create(
            celery_task_id='random id',
            name='error_task',
            queue_name='default',
            input='',
            task_args=[],
            task_kwargs={},
            estimated_time_of_first_arrival=now() - timedelta(minutes=5),
            expires=now() - timedelta(minutes=4),
            stale=now() - timedelta(minutes=3)
        )
        test_call_command('set_celery_task_log_state')
        assert_equal(celery_task_log.refresh_from_db().state, CeleryTaskLogState.EXPIRED)
        assert_false(CeleryTaskLog.objects.filter_processing().exists())

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_stale_failed_celery_task_should_not_be_set_as_failed_with_command(self):
        sum_task.apply_async(args=(5, 8))
        celery_task_log = CeleryTaskLog.objects.create(
            celery_task_id='random id',
            name='error_task',
            queue_name='default',
            input='',
            task_args=[],
            task_kwargs={},
            estimated_time_of_first_arrival=now() - timedelta(minutes=5),
            expires=now() - timedelta(minutes=4),
            stale=now() - timedelta(minutes=3)
        )
        CeleryTaskRunLog.objects.create(
            celery_task_id='random id',
            name='error_task',
            state=CeleryTaskRunLogState.FAILED,
            task_args=[],
            task_kwargs={},
            retries=0
        )

        test_call_command('set_celery_task_log_state')
        assert_equal(celery_task_log.refresh_from_db().state, CeleryTaskLogState.FAILED)

    @freeze_time(now())
    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, CELERYD_TASK_TIME_LIMIT=10)
    def test_stale_succeded_celery_task_should_not_be_set_as_succeeded_with_command(self):
        sum_task.apply_async(args=(5, 8))
        celery_task_log = CeleryTaskLog.objects.create(
            celery_task_id='random id',
            name='error_task',
            queue_name='default',
            input='',
            task_args=[],
            task_kwargs={},
            estimated_time_of_first_arrival=now() - timedelta(minutes=5),
            expires=now() - timedelta(minutes=4),
            stale=now() - timedelta(minutes=3)
        )
        CeleryTaskRunLog.objects.create(
            celery_task_id='random id',
            name='error_task',
            state=CeleryTaskRunLogState.SUCCEEDED,
            task_args=[],
            task_kwargs={},
            retries=0
        )

        test_call_command('set_celery_task_log_state')
        assert_equal(celery_task_log.refresh_from_db().state, CeleryTaskLogState.SUCCEEDED)

    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=None)
    def test_unique_task_shoud_have_set_stale_limit(self):
        with assert_raises(CeleryError):
            unique_task.delay()
        with override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=10):
            with assert_not_raises(CeleryError):
                unique_task.delay()

    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=5)
    def test_apply_async_and_get_result_should_return_time_error_for_zero_timeout(self):
        with assert_raises(TimeoutError):
            unique_task.apply_async_and_get_result(timeout=0)

    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=5)
    def test_apply_async_and_get_result_should_return_task_result(self):
        assert_equal(unique_task.apply_async_and_get_result(), 'unique')

    @responses.activate
    def test_output_logged_request_should_be_related_with_object_selected_in_decorator(self):
        user1 = self.create_user('test1', 'test1@test.cz')
        user2 = self.create_user('test2', 'test2@test.cz')

        responses.add(responses.GET, 'http://test.cz', body='test')
        assert_equal(log_context_manager.get_output_request_related_objects(), [])
        with atomic_log(output_requests_slug='test1', output_requests_related_objects=[user1]):
            assert_equal(set(log_context_manager.get_output_request_related_objects()), {user1})

            requests.get('http://test.cz')
            with atomic_log(output_requests_slug='test2', output_requests_related_objects=[user2]):
                assert_equal(set(log_context_manager.get_output_request_related_objects()), {user1, user2})
                requests.get('http://test.cz')
            assert_equal(set(log_context_manager.get_output_request_related_objects()), {user1})
            requests.get('http://test.cz')

        assert_equal(OutputLoggedRequest.objects.count(), 3)
        logged_request3, logged_request2, logged_request1 = OutputLoggedRequest.objects.all()

        assert_equal(logged_request1.related_objects.get().object, user1)
        assert_equal(logged_request1.slug, 'test1')

        assert_equal(
            {
                related_object.object
                for related_object in logged_request2.related_objects.all()
            },
            {user1, user2}
        )
        assert_equal(logged_request2.slug, 'test2')

        assert_equal(logged_request3.related_objects.get().object, user1)
        assert_equal(logged_request3.slug, 'test1')
        assert_equal(log_context_manager.get_output_request_related_objects(), [])
        assert_is_none(log_context_manager.get_output_request_slug())

    @data_consumer('create_user')
    def test_task_log_is_processing_should_return_if_task_is_active_or_waiting(self, user):
        assert_false(sum_task.is_processing())

        celery_task_log = CeleryTaskLog.objects.create(
            celery_task_id='test',
            name='sum_task',
            stale=now(),
            estimated_time_of_first_arrival=now()
        )
        celery_task_log.related_objects.add(user)

        assert_true(sum_task.is_processing())
        assert_true(sum_task.is_processing(related_objects=[user]))

        user2 = self.create_user(username='test2', email='test2@test.cz')
        assert_false(sum_task.is_processing(related_objects=(user2,)))

        CeleryTaskRunLog(
            celery_task_id='test',
            name='sum_task'
        )
        assert_true(sum_task.is_processing())
        assert_true(sum_task.is_processing(related_objects=[user]))

        celery_task_log.change_and_save(state=CeleryTaskLogState.SUCCEEDED)
        assert_false(sum_task.is_processing())

        celery_task_log.change_and_save(state=CeleryTaskLogState.FAILED)
        assert_false(sum_task.is_processing())

        celery_task_log.change_and_save(state=CeleryTaskLogState.RETRIED)
        assert_true(sum_task.is_processing())

        celery_task_log.change_and_save(state=CeleryTaskLogState.EXPIRED)
        assert_false(sum_task.is_processing())

        celery_task_log.change_and_save(state=CeleryTaskLogState.ACTIVE)
        assert_true(sum_task.is_processing())

        celery_task_log.change_and_save(state=CeleryTaskLogState.EXPIRED)
        assert_false(sum_task.is_processing())


class CeleryHealthCheckCommandTestCase(CeleryHealthCheckCommandTestCaseMixin, GermaniumTestCase):

    @data_consumer('create_waiting_celery_task_log')
    def test_celery_health_check_command_raises_error_if_queue_exceeds_max_tasks_count_limit(self, task_log):
        max_tasks_count = 0

        with assert_raises(CommandError):
            call_command('celery_health_check', max_tasks_count=max_tasks_count)

    @data_consumer('create_waiting_celery_task_log')
    def test_celery_health_check_command_does_not_raise_error_if_queue_is_within_max_tasks_count_limit(self, task_log):
        with assert_not_raises(CommandError):
            call_command('celery_health_check', max_tasks_count=1)

    def test_celery_health_check_command_raises_command_error_for_max_tasks_count_lower_than_zero(self):
        with assert_raises(CommandError):
            call_command('celery_health_check', max_tasks_count=-1)

    def test_celery_health_check_command_raises_command_error_for_max_created_at_diff_lower_than_zero(self):
        with assert_raises(CommandError):
            call_command('celery_health_check', max_created_at_diff=-1)

    @freeze_time(now())
    @data_consumer('create_waiting_celery_task_log')
    def test_celery_health_check_command_does_not_raise_error_if_queue_is_within_max_created_at_diff_limit(self,
                                                                                                           task_log):
        with assert_not_raises(CommandError):
            call_command('celery_health_check', max_created_at_diff=5)

    @freeze_time(now() - timedelta(days=1))
    @data_consumer('create_waiting_celery_task_log')
    def test_celery_health_check_command_does_raise_error_if_queue_is_not_within_max_created_at_diff_limit(self,
                                                                                                           task_log):
        max_created_at_diff = 1

        with assert_raises(CommandError):
            call_command('celery_health_check', max_created_at_diff=max_created_at_diff)



from django.core.management.base import BaseCommand, CommandError
from django.db.models import DurationField, ExpressionWrapper, F
from django.db.models.functions import Now
from django.utils.timezone import timedelta

from security.config import settings
from security.models import CeleryTaskLog, CeleryTaskRunLog


class Command(BaseCommand):

    help = 'Check number of waiting Celery tasks in given queue'
    messages = {
        'invalid_max_created_at_diff': 'Max created at diff must be >= 0.',
        'invalid_max_tasks_count': 'Max tasks count must be >= 0.',
        'max_created_at_diff_error': ('There is a task waiting for more than {} seconds in queue "{}" and therefore it '
                                      'is considered as overfilled.'),
        'max_tasks_count_error': ('There are more than {} waiting tasks in queue "{}" and therefore it is considered as'
                                  'overfilled.'),
        'missing_options_error': 'Specify either "--max-created-at-diff" or "--max-tasks-count".',
        'ok': 'Queue "{}" is ok.',
    }

    def add_arguments(self, parser):
        parser.add_argument(
            '-q', '--queue-name',
            default=settings.CELERY_HEALTH_CHECK_DEFAULT_QUEUE,
            dest='queue_name',
            help='Default queue name.',
            type=str,
        )
        parser.add_argument(
            '-d', '--max-created-at-diff',
            dest='max_created_at_diff',
            help='Max created at difference (in seconds).',
            type=int,
        )
        parser.add_argument(
            '-t', '--max-tasks-count',
            dest='max_tasks_count',
            help='Max count of waiting tasks.',
            type=int,
        )

    def _get_queryset(self, queue_name):
        return CeleryTaskLog.objects.filter_waiting().order_by('created_at')

    def _handle_max_created_at_diff(self, queryset, queue_name, max_created_at_diff):
        if max_created_at_diff < 0:
            raise CommandError(self.messages['invalid_max_created_at_diff'])

        if queryset.annotate(diff=ExpressionWrapper(Now() - F('created_at'), output_field=DurationField())).filter(
            diff__gt=timedelta(seconds=max_created_at_diff),
        ).exists():
            raise CommandError(self.messages['max_created_at_diff_error'].format(max_created_at_diff, queue_name))

    def _handle_max_tasks(self, queryset, queue_name, max_tasks_count):
        if max_tasks_count < 0:
            raise CommandError(self.messages['invalid_max_tasks_count'])

        if queryset.count() > max_tasks_count:
            raise CommandError(self.messages['max_tasks_count_error'].format(max_tasks_count, queue_name))

    def handle(self, **options):
        max_created_at_diff = options.get('max_created_at_diff')
        max_tasks_count = options.get('max_tasks_count')

        if max_created_at_diff is None and max_tasks_count is None:
            raise CommandError(self.messages['missing_options_error'])

        queue_name = options['queue_name']
        queryset = self._get_queryset(queue_name)

        try:
            if max_created_at_diff is not None:
                self._handle_max_created_at_diff(queryset, queue_name, max_created_at_diff)
            if max_tasks_count is not None:
                self._handle_max_tasks(queryset, queue_name, max_tasks_count)
        except CommandError as ex:
            raise ex
        else:
            self.stdout.write(self.style.SUCCESS(self.messages['ok'].format(queue_name)))



from django.conf import settings as django_settings

from attrdict import AttrDict


DEFAULTS = {
    'DEFAULT_THROTTLING_VALIDATORS_PATH': 'security.default_validators',
    'THROTTLING_FAILURE_VIEW': 'security.views.throttling_failure_view',
    'LOG_REQUEST_IGNORE_IP': (),
    'LOG_REQUEST_IGNORE_URL_PATHS': (),
    'LOG_REQUEST_BODY_LENGTH': 1000,
    'LOG_RESPONSE_BODY_LENGTH': 1000,
    'LOG_RESPONSE_BODY_CONTENT_TYPES': (
        'application/json', 'application/xml', 'text/xml', 'text/csv', 'text/html', 'application/xhtml+xml'
    ),
    'LOG_JSON_STRING_LENGTH': 250,
    'COMMAND_LOG_EXCLUDED_COMMANDS': (
        'runserver', 'makemigrations', 'migrate', 'sqlmigrate', 'showmigrations', 'shell', 'shell_plus', 'test',
        'help', 'reset_db', 'compilemessages', 'makemessages', 'dumpdata', 'loaddata'
    ),
    'HIDE_SENSITIVE_DATA': True,
    'HIDE_SENSITIVE_DATA_PATTERNS': {
        'BODY': (
            r'"password"\s*:\s*"((?:\\"|[^"])*)',
            r'<password>([^<]*)',
            r'password=([^&]*)',
            r'csrfmiddlewaretoken=([^&]*)',
            r'(?i)content-disposition: form-data; name="password"\r\n\r\n.*',
            r'"access_key": "([^"]*)',
        ),
        'HEADERS': (
            r'Authorization',
            r'X_Authorization',
            r'Cookie',
            r'.*token.*',
        ),
        'QUERIES': (
            r'.*token.*',
        ),
    },
    'SENSITIVE_DATA_REPLACEMENT': '[Filtered]',
    'APPEND_SLASH': True,
    'BACKUP_STORAGE_CLASS': 'security.storages.BackupFileSystemStorage',
    'BACKUP_STORAGE_PATH': None,
    'DEBUG_TOOLBAR': False,
    'SHOW_DEBUG_TOOLBAR': False,
    'LOG_OUTPUT_REQUESTS': True,
    'AUTO_GENERATE_TASKS_FOR_DJANGO_COMMANDS': {},
    'CELERY_HEALTH_CHECK_DEFAULT_QUEUE': getattr(django_settings, 'CELERY_DEFAULT_QUEUE', 'default'),
    'LOG_DB_NAME': None,
    'PURGE_LOG_BACKUP_BATCH': 20000,
    'PURGE_LOG_DELETE_BATCH': 1000,
    'TASK_LOGGER_NAME': 'security.task',
}


class Settings:

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid Security setting: "{}"'.format(attr))

        value = getattr(django_settings, 'SECURITY_{}'.format(attr), DEFAULTS[attr])

        if isinstance(value, dict):
            value = AttrDict(value)

        return value


settings = Settings()



# Generated by Django 2.2.16 on 2020-10-07 12:03

from django.db import migrations, models
import django.db.models.deletion
import enumfields.fields
import security.enums


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTaskLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('celery_task_id', models.CharField(db_index=True, max_length=250, verbose_name='celery ID')),
                ('name', models.CharField(db_index=True, max_length=250, verbose_name='task name')),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name')),
                ('input', models.TextField(blank=True, editable=False, null=True, verbose_name='input')),
                ('task_args',
                 models.TextField(blank=True, editable=False, null=True, verbose_name='task args')),
                ('task_kwargs',
                 models.TextField(blank=True, editable=False, null=True, verbose_name='task kwargs')),
                ('estimated_time_of_first_arrival',
                 models.DateTimeField(verbose_name='estimated time of first arrival')),
                ('expires', models.DateTimeField(blank=True, null=True, verbose_name='time of expiration')),
                ('stale', models.DateTimeField(blank=True, null=True, verbose_name='stale task time')),
                ('state',
                 enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CeleryTaskLogState,
                                                verbose_name='state')),
            ],
            options={
                'verbose_name': 'celery task',
                'verbose_name_plural': 'celery tasks',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('celery_task_id', models.CharField(db_index=True, max_length=250, verbose_name='celery ID')),
                ('start', models.DateTimeField(blank=True, null=True, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='time')),
                ('state',
                 enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CeleryTaskRunLogState,
                                                verbose_name='state')),
                ('name', models.CharField(db_index=True, max_length=250, verbose_name='task name')),
                ('task_args',
                 models.TextField(blank=True, editable=False, null=True, verbose_name='task args')),
                ('task_kwargs',
                 models.TextField(blank=True, editable=False, null=True, verbose_name='task kwargs')),
                ('result', models.TextField(blank=True, editable=False, null=True, verbose_name='result')),
                ('error_message', models.TextField(blank=True, null=True, verbose_name='error message')),
                ('output', models.TextField(blank=True, editable=False, null=True, verbose_name='output')),
                ('retries', models.PositiveSmallIntegerField(default=0, verbose_name='retries')),
                ('estimated_time_of_next_retry',
                 models.DateTimeField(blank=True, null=True, verbose_name='estimated time of arrival')),
            ],
            options={
                'verbose_name': 'celery task run',
                'verbose_name_plural': 'celery tasks run',
                'ordering': ('created_at',),
            },
        ),
        migrations.CreateModel(
            name='CommandLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='time')),
                ('name', models.CharField(db_index=True, editable=False, max_length=250, verbose_name='name')),
                ('input', models.TextField(editable=False, verbose_name='input')),
                ('executed_from_command_line',
                 models.BooleanField(default=False, editable=False, verbose_name='executed from command line')),
                ('output', models.TextField(blank=True, editable=False, null=True, verbose_name='output')),
                ('is_successful',
                 models.BooleanField(default=False, editable=False, verbose_name='finished successfully')),
                (
                'error_message', models.TextField(blank=True, editable=False, null=True, verbose_name='error message')),
            ],
            options={
                'verbose_name': 'command log',
                'verbose_name_plural': 'command logs',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='InputLoggedRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('host', models.CharField(db_index=True, max_length=255, verbose_name='host')),
                ('method', models.SlugField(max_length=7, verbose_name='method')),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000, verbose_name='URL path')),
                ('queries', models.TextField(blank=True, null=True, verbose_name='queries')),
                ('is_secure', models.BooleanField(default=False, verbose_name='HTTPS connection')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('request_timestamp', models.DateTimeField(db_index=True, verbose_name='request timestamp')),
                ('request_headers', models.TextField(blank=True, null=True, verbose_name='request headers')),
                ('request_body', models.TextField(blank=True, verbose_name='request body')),
                ('response_timestamp',
                 models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='response timestamp')),
                ('response_code',
                 models.PositiveSmallIntegerField(blank=True, null=True, verbose_name='response code')),
                ('response_headers', models.TextField(blank=True, null=True, verbose_name='response headers')),
                ('response_body', models.TextField(blank=True, null=True, verbose_name='response body')),
                ('response_time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('status', enumfields.fields.NumEnumField(default=0, enum=security.enums.LoggedRequestStatus,
                                                          verbose_name='status')),
                ('error_description', models.TextField(blank=True, null=True, verbose_name='error description')),
                ('exception_name',
                 models.CharField(blank=True, max_length=255, null=True, verbose_name='exception name')),
                ('user_id', models.TextField(blank=True, db_index=True, null=True, verbose_name='user ID')),
                ('ip', models.GenericIPAddressField(db_index=True, verbose_name='IP address')),
                ('type',
                 enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.InputLoggedRequestType,
                                                verbose_name='type')),
            ],
            options={
                'verbose_name': 'input logged request',
                'verbose_name_plural': 'input logged requests',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputLoggedRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('host', models.CharField(db_index=True, max_length=255, verbose_name='host')),
                ('method', models.SlugField(max_length=7, verbose_name='method')),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000, verbose_name='URL path')),
                ('queries', models.TextField(blank=True, null=True, verbose_name='queries')),
                ('is_secure', models.BooleanField(default=False, verbose_name='HTTPS connection')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('request_timestamp', models.DateTimeField(db_index=True, verbose_name='request timestamp')),
                ('request_headers', models.TextField(blank=True, null=True, verbose_name='request headers')),
                ('request_body', models.TextField(blank=True, verbose_name='request body')),
                ('response_timestamp',
                 models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='response timestamp')),
                ('response_code', models.PositiveSmallIntegerField(blank=True, null=True,
                                                                   verbose_name='response code')),
                ('response_headers', models.TextField(blank=True, null=True, verbose_name='response headers')),
                ('response_body', models.TextField(blank=True, null=True, verbose_name='response body')),
                ('response_time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('status', enumfields.fields.NumEnumField(default=0, enum=security.enums.LoggedRequestStatus,
                                                          verbose_name='status')),
                ('error_description', models.TextField(blank=True, null=True, verbose_name='error description')),
                ('exception_name',
                 models.CharField(blank=True, max_length=255, null=True, verbose_name='exception name')),
            ],
            options={
                'verbose_name': 'output logged request',
                'verbose_name_plural': 'output logged requests',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputLoggedRequestGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('output_logged_request',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.OutputLoggedRequest')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('output_logged_request', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='InputLoggedRequestGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('input_logged_request',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.InputLoggedRequest')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('input_logged_request', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CommandLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('command_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.CommandLog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('command_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('celery_task_run_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.CeleryTaskRunLog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_run_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('celery_task_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.CeleryTaskLog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_log', 'object_ct_id', 'object_id')},
            },
        ),
    ]



from settings.celery import app as celery_app

from security.task import LoggedTask


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='sum_task')
def sum_task(self, a, b):
    return a + b


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='error_task',
    stale_time_limit=60 * 60)
def error_task(self):
    raise RuntimeError('error')


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='retry_task',
    autoretry_for=(RuntimeError,),
    default_retry_delays=(1 * 60, 5 * 60, 10 * 60, 30 * 60, 60 * 60))
def retry_task(self):
    if self.request.retries != 5:
        raise RuntimeError('error')


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='unique_task',
    unique=True)
def unique_task(self):
    return 'unique'



from django.db import models
from django.utils.translation import ugettext_lazy as _

from security.models import InputLoggedRequest


class DebugToolbarData(models.Model):

    logged_request = models.OneToOneField(
        InputLoggedRequest,
        verbose_name=_('logged request'),
        null=False,
        blank=False,
        on_delete=models.CASCADE,
        related_name='input_logged_request_toolbar'
    )
    toolbar = models.TextField(
        verbose_name=_('toolbar'),
        null=False,
        blank=False,
    )

    def __str__(self):
        return ' #%s' % self.pk

    class Meta:
        verbose_name = _('Logged request toolbar')
        verbose_name_plural = _('Logged requests toolbars')



from django.utils.translation import ugettext_lazy as _

from enumfields import Choice, ChoiceEnum


class LoggedRequestStatus(ChoiceEnum):

    INCOMPLETE = Choice(0, _('Incomplete'))
    DEBUG = Choice(10, _('Debug'))
    INFO = Choice(20, _('Info'))
    WARNING = Choice(30, _('Warning'))
    ERROR = Choice(40, _('Error'))
    CRITICAL = Choice(50, _('Critical'))


class InputLoggedRequestType(ChoiceEnum):

    COMMON_REQUEST = Choice(1, _('Common request'))
    THROTTLED_REQUEST = Choice(2, _('Throttled request'))
    SUCCESSFUL_LOGIN_REQUEST = Choice(3, _('Successful login request'))
    UNSUCCESSFUL_LOGIN_REQUEST = Choice(4, _('Unsuccessful login request'))
    SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = Choice(5, _('Successful two factor code verification request'))
    UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = Choice(6, _('Unsuccessful two factor code verification request'))


class CeleryTaskLogState(ChoiceEnum):

    WAITING = Choice(1, _('Waiting'))
    ACTIVE = Choice(2, _('Active'))
    SUCCEEDED = Choice(3, _('Succeeded'))
    FAILED = Choice(4, _('Failed'))
    RETRIED = Choice(5, _('Retried'))
    EXPIRED = Choice(6, _('Expired'))


class CeleryTaskRunLogState(ChoiceEnum):

    ACTIVE = Choice(1, _('Active'))
    SUCCEEDED = Choice(2, _('Succeeded'))
    FAILED = Choice(3, _('Failed'))
    RETRIED = Choice(4, _('Retried'))
    EXPIRED = Choice(5, _('Expired'))



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0001_migration'),
    ]

    operations = [
        migrations.CreateModel(
            name='DebugToolbarData',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('logged_request',models.OneToOneField(
                    on_delete=models.deletion.CASCADE, related_name='input_logged_request_toolbar',
                    to='security.InputLoggedRequest', verbose_name='logged request'
                )),
                ('toolbar',  models.TextField(verbose_name='toolbar', null=False, blank=False,)),
            ],
            options={
                'verbose_name': 'Logged request toolbar',
                'verbose_name_plural': 'Logged requests toolbars',
            },
        ),
    ]



import itertools
from contextlib import ContextDecorator
from threading import local

from security.models import OutputLoggedRequest


_input_logged_request = local()
_output_logged_requests_list = local()


class OutputLoggedRequestContext:
    """
    Data structure that stores data for creation OutputLoggedRequest model object
    """

    def __init__(self, data, related_objects=None):
        self.data = data
        self.related_objects = related_objects

    def create(self, input_logged_request=None):
        output_logged_request = OutputLoggedRequest.objects.create(
            input_logged_request=input_logged_request, **self.data
        )
        output_logged_request.related_objects.add(*self.related_objects)
        return output_logged_request


class AtomicLog(ContextDecorator):
    """
    Context decorator that stores logged requests to database connections and inside exit method
    stores it to the database
    """

    def __enter__(self):
        output_logged_requests_list = getattr(_output_logged_requests_list, 'value', [])
        output_logged_requests_list.append([])
        _output_logged_requests_list.value = output_logged_requests_list

    def __exit__(self, exc_type, exc_value, traceback):
        output_logged_requests = _output_logged_requests_list.value.pop()
        if _output_logged_requests_list.value:
            _output_logged_requests_list.value[-1] += output_logged_requests
        else:
            [output_logged_request.create() for output_logged_request in output_logged_requests]


def atomic_log(function=None):
    """
    Decorator that surrounds atomic block, ensures that logged output requests will be stored inside database in case
    of DB rollback
    """
    if callable(function):
        return AtomicLog()(function)
    else:
        return AtomicLog()


def is_active_logged_requests():
    """
    :param using: database alias
    :return: True if block of code is surrounded with atomic_log operator
    """
    return hasattr(_output_logged_requests_list, 'value') and _output_logged_requests_list.value


def get_all_request_related_objects():
    """
    Return all related selected with log_request_related_objects decorator.
    :return: list of request related objects
    """
    related_objects_list = getattr(_input_logged_request, 'related_objects_list', [])
    return list(itertools.chain(*related_objects_list))


def get_request_slug():
    """
    Return last slug defined with log_request_related_objects decorator.
    :return input logged request slug
    """
    slug_list = getattr(_input_logged_request, 'slug_list', [])
    return None if not slug_list else slug_list[-1]


class LogRequestRelatedObjects(ContextDecorator):
    """
    Context decorator that adds related objects to all requests inside code block
    """

    def __init__(self, slug, related_objects):
        self.slug = slug
        self.related_objects = [] if related_objects is None else related_objects

    def __enter__(self):
        related_objects_list = getattr(_input_logged_request, 'related_objects_list', [])
        slug_list = getattr(_input_logged_request, 'slug_list', [])

        last_related_objects_slug = None if not slug_list else slug_list[-1]

        slug_list.append(last_related_objects_slug if self.slug is None else self.slug)
        related_objects_list.append(self.related_objects)
        _input_logged_request.related_objects_list = related_objects_list
        _input_logged_request.slug_list = slug_list

    def __exit__(self, exc_type, exc_value, traceback):
        _input_logged_request.related_objects_list.pop()
        _input_logged_request.slug_list.pop()


def log_request_related_objects(slug=None, related_objects=None):
    """
    Decorator that adds related objects to all requests inside code block.
    """
    return LogRequestRelatedObjects(slug, related_objects)



import os
import base64
import logging
import pickle

from datetime import timedelta

from distutils.version import StrictVersion

from django.conf import settings as django_settings
from django.core.management import call_command, get_commands, load_command_class
from django.core.management.base import OutputWrapper
from django.core.exceptions import ImproperlyConfigured
from django.db import close_old_connections, transaction
from django.db.utils import InterfaceError, OperationalError
from django.utils.timezone import now

try:
    from celery import Task, shared_task, current_app
    from celery.result import AsyncResult
    from celery.exceptions import CeleryError, TimeoutError
    from celery.worker.request import Request
except ImportError:
    raise ImproperlyConfigured('Missing celery library, please install it')

from django_celery_extensions.task import DjangoTask

from .decorators import atomic_log
from .config import settings
from .models import CeleryTaskRunLog, CeleryTaskLog, CeleryTaskLogState, CeleryTaskRunLogState
from .utils import LogStringIO, log_context_manager


logger = logging.getLogger(settings.TASK_LOGGER_NAME)


class LoggedTask(DjangoTask):

    abstract = True

    _log_messages = {
        'expire': 'Task "%(task_name)s" (%(task)s) is expired',
        'failure': 'Task "%(task_name)s" (%(task)s) is failed',
        'retry': 'Task "%(task_name)s" (%(task)s) is retried',
        'success': 'Task "%(task_name)s" (%(task)s) is completed',
        'apply': 'Task "%(task_name)s" (%(task)s) is applied',
        'start': 'Task "%(task_name)s" (%(task)s) is running',
    }

    def _get_log_message(self, log_name):
        return self._log_messages[log_name]

    def _store_log_output(self, output_stream):
        self.request.celery_task_run_log.change_and_save(
            output=output_stream.getvalue(),
            update_only_changed_fields=True
        )

    @property
    def stdout(self):
        return OutputWrapper(self.request.output_stream)

    @property
    def stderr(self):
        return OutputWrapper(self.request.output_stream)

    @property
    def task_run_log(self):
        return self.request.celery_task_run_log

    @property
    def task_log(self):
        return self.request.celery_task_log

    def on_start(self, args, kwargs):
        super().on_start(args, kwargs)
        celery_task_run_log = self._create_task_run_log(args, kwargs)
        # Every set attr is sent here
        self.request.celery_task_run_log = celery_task_run_log
        self.request.output_stream = LogStringIO(post_write_callback=self._store_log_output)
        self.on_start_task(self.task_run_log, args, kwargs)

    def _start(self, *args, **kwargs):
        with atomic_log(celery_task_run_log=self.request.celery_task_run_log):
            return super()._start(*args, **kwargs)

    def on_apply(self, task_id, apply_time, stale_time_limit, args, kwargs, options):
        super().on_apply(task_id, apply_time, stale_time_limit, args, kwargs, options)
        task_initiation = self._create_task_log(
            task_id,
            args,
            kwargs,
            apply_time,
            options.get('eta'),
            options.get('expires'),
            stale_time_limit,
            options.get('queue'),
            options.pop('related_objects', None)
        )
        self.on_apply_task(task_initiation, args, kwargs, options)

    def on_apply_task(self, task_log, args, kwargs, options):
        """
        On apply task is invoked before task was prepared. Therefore task request context is not prepared.
        :param task_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param options: input task options
        """
        logger.info(
            self._get_log_message('apply'),
            dict(
                task=task_log.celery_task_id,
                task_name=task_log.name,
                **(kwargs or {})
            ),
            extra=dict(
                task_args=args,
                task_id=task_log.celery_task_id,
                task_input=task_log.input,
                task_kwargs=kwargs,
                task_log_id=task_log.pk,
                task_name=task_log.name,
                task_queue=task_log.queue_name,
                task_state=task_log.state.name,
            ),
        )

    def on_apply_retry(self, args, kwargs, exc, eta):
        super().on_apply_retry(args, kwargs, exc, eta)
        self.on_retry_task(self.task_run_log, args, kwargs, exc, eta)

    def on_retry_task(self, task_run_log, args, kwargs, exc, eta):
        """
        On retry task is invoked before task was retried.
        :param task_run_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param exc: raised exception which caused retry
        """
        task_log = task_run_log.get_task_log()
        logger.warning(
            self._get_log_message('retry'),
            dict(
                attempt=self.request.retries,
                exception=str(exc),
                task=task_run_log.celery_task_id,
                task_name=task_run_log.name,
                **(kwargs or {})
            ),
            extra=dict(
                task_args=args,
                task_exception=str(exc),
                task_id=task_run_log.celery_task_id,
                task_kwargs=kwargs,
                task_log_id=None if not task_log else task_log.pk,
                task_name=task_run_log.name,
                task_queue=None if not task_log else task_log.queue_name,
                task_retries=self.request.retries,
                task_run_log_id=task_run_log.pk,
                task_started_at=task_run_log.start,
                task_state=task_run_log.state.name,
            )
        )
        stop = now()
        task_run_log.change_and_save(
            state=CeleryTaskLogState.RETRIED,
            error_message=str(exc),
            stop=stop,
            time=(stop - task_run_log.start).total_seconds(),
            output=self.request.output_stream.getvalue(),
            estimated_time_of_next_retry=eta
        )
        task_log = task_run_log.get_task_log()
        if task_log:
            task_log.change_and_save(
                state=CeleryTaskLogState.RETRIED
            )

    def on_start_task(self, task_run_log, args, kwargs):
        """
        On start task is invoked during task started.
        :param task_run_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        """
        task_log = task_run_log.get_task_log()
        logger.info(
            self._get_log_message('start'),
            dict(
                attempt=self.request.retries,
                task=task_run_log.celery_task_id,
                task_name=task_run_log.name,
                **(kwargs or {})
            ),
            extra=dict(
                task_args=args,
                task_id=task_run_log.celery_task_id,
                task_kwargs=kwargs,
                task_log_id=None if not task_log else task_log.pk,
                task_name=task_run_log.name,
                task_queue=None if not task_log else task_log.queue_name,
                task_run_log_id=task_run_log.pk,
                task_started_at=task_run_log.start,
                task_state=task_run_log.state.name,
            )
        )
        if task_log:
            task_log.change_and_save(
                state=CeleryTaskLogState.ACTIVE
            )

    def on_success_task(self, task_run_log, args, kwargs, retval):
        """
        On start task is invoked if task is successful.
        :param task_run_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param retval: return value of a task
        """
        task_log = task_run_log.get_task_log()
        logger.info(
            self._get_log_message('success'),
            dict(
                attempt=self.request.retries,
                task=task_run_log.celery_task_id,
                task_name=task_run_log.name,
                **(kwargs or {})
            ),
            extra=dict(
                task_args=args,
                task_duration=task_run_log.time,
                task_empty_error=task_run_log.error_message is None or len(task_run_log.error_message) < 1,
                task_empty_output=task_run_log.output is None or len(task_run_log.output) < 1,
                task_id=task_run_log.celery_task_id,
                task_kwargs=kwargs,
                task_log_id=None if not task_log else task_log.pk,
                task_name=task_run_log.name,
                task_queue=None if not task_log else task_log.queue_name,
                task_retries=task_run_log.retries,
                task_retval=retval,
                task_run_log_id=task_run_log.pk,
                task_started_at=task_run_log.start,
                task_state=task_run_log.state.name,
            )
        )

        if retval:
            self.request.output_stream.write('Return value is "{}"'.format(retval))

        stop = now()
        task_run_log.change_and_save(
            state=CeleryTaskRunLogState.SUCCEEDED,
            stop=stop,
            time=(stop - task_run_log.start).total_seconds(),
            result=retval,
            output=self.request.output_stream.getvalue()
        )

        if task_log:
            task_log.change_and_save(
                state=CeleryTaskLogState.SUCCEEDED
            )

    def on_success(self, retval, task_id, args, kwargs):
        super().on_success(retval, task_id, args, kwargs)
        self.on_success_task(self.task_run_log, args, kwargs, retval)

    def on_failure_task(self, task_run_log, args, kwargs, exc):
        """
        On failure task is invoked if task end with some exception.
        :param task_run_log: logged celery task instance
        :param args: task args
        :param kwargs: task kwargs
        :param exc: raised exception which caused retry
        """
        task_log = task_run_log.get_task_log()
        logger.error(
            self._get_log_message('failure'),
            dict(
                attempt=self.request.retries,
                exception=str(exc),
                task=task_run_log.celery_task_id,
                task_name=task_run_log.name,
                **(kwargs or {})
            ),
            extra=dict(
                task_args=args,
                task_duration=task_run_log.time,
                task_empty_error=task_run_log.error_message is None or len(task_run_log.error_message) < 1,
                task_empty_output=task_run_log.output is None or len(task_run_log.output) < 1,
                task_exception=str(exc),
                task_id=task_run_log.celery_task_id,
                task_kwargs=kwargs,
                task_log_id=None if not task_log else task_log.pk,
                task_name=task_run_log.name,
                task_queue=None if not task_log else task_log.queue_name,
                task_retries=task_run_log.retries,
                task_run_log_id=task_run_log.pk,
                task_started_at=task_run_log.start,
                task_state=task_run_log.state.name,
            )
        )
        stop = now()
        task_run_log.change_and_save(
            state=CeleryTaskRunLogState.FAILED,
            stop=stop,
            time=(stop - task_run_log.start).total_seconds(),
            error_message=str(exc),
            output=self.request.output_stream.getvalue()
        )
        if task_log:
            task_log.change_and_save(
                state=CeleryTaskLogState.FAILED
            )

    def expire_task(self, task_log):
        logger.error(
            self._get_log_message('expire'),
            dict(
                task=task_log.celery_task_id,
                task_name=task_log.name
            ),
            extra=dict(
                task_expired_at=task_log.stale,
                task_id=task_log.celery_task_id,
                task_input=task_log.input,
                task_log_id=task_log.pk,
                task_name=task_log.name,
                task_queue=task_log.queue_name,
                task_state=task_log.state.name,
            )
        )
        task_log.change_and_save(
            state=CeleryTaskLogState.EXPIRED
        )
        task_log.runs.filter(
            state=CeleryTaskRunLogState.ACTIVE
        ).update(
            state=CeleryTaskRunLogState.EXPIRED
        )

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        super().on_failure(exc, task_id, args, kwargs, einfo)
        try:
            self.on_failure_task(self.task_run_log, args, kwargs, exc)
        except CeleryTaskRunLog.DoesNotExist:
            pass

    def _create_task_log(self, task_id, task_args, task_kwargs, apply_time, eta, expires, stale_time_limit,
                         queue, related_objects):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]

        celery_task_log = CeleryTaskLog.objects.create(
            celery_task_id=task_id,
            name=self.name,
            queue_name=queue,
            input=', '.join(task_input),
            task_args=task_args,
            task_kwargs=task_kwargs,
            estimated_time_of_first_arrival=eta,
            expires=expires,
            stale=apply_time + timedelta(seconds=stale_time_limit) if stale_time_limit is not None else None
        )
        related_objects = [] if related_objects is None else list(related_objects)
        if log_context_manager.is_active():
            related_objects += log_context_manager.get_logs()

        celery_task_log.related_objects.add(*related_objects)
        return celery_task_log

    def _create_task_run_log(self, task_args, task_kwargs):
        return CeleryTaskRunLog.objects.create(
            celery_task_id=self.request.id,
            name=self.name,
            task_args=task_args,
            task_kwargs=task_kwargs,
            retries=self.request.retries,
            state=CeleryTaskRunLogState.ACTIVE,
            start=now(),
        )

    def is_processing(self, related_objects=None):
        return CeleryTaskLog.objects.filter_processing(
            name=self.name,
            related_objects=related_objects
        ).exists()

    def get_command_kwargs(self):
        return dict(
            stdout=self.request.output_stream,
            stderr=self.request.output_stream,
        )



from django.contrib import admin

from apps.test_security.views import (
    extra_throttling_view, hide_request_body_view, log_exempt_view, proxy_view, throttling_exempt_view
)

try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url


urlpatterns = [
    url('admin/', admin.site.urls),
    url('proxy/', proxy_view),
    url('hide-request-body/', hide_request_body_view),
    url('log-exempt/', log_exempt_view),
    url('throttling-exempt/', throttling_exempt_view),
    url('extra-throttling/', extra_throttling_view),
]



from django.conf import settings as django_settings

from debug_toolbar.toolbar import DebugToolbar

from security.config import settings

from .models import DebugToolbarData


class DebugToolbarLogMiddleware:

    def __init__(self, get_response=None):
        self.get_response = get_response
        super().__init__()

    def __call__(self, request):
        if settings.DEBUG_TOOLBAR and django_settings.DEBUG:
            toolbar = DebugToolbar(request, self.get_response)

            # Activate instrumentation ie. monkey-patch.
            for panel in toolbar.enabled_panels:
                panel.enable_instrumentation()
            try:
                # Run panels like Django middleware.
                response = toolbar.process_request(request)
            finally:
                # Deactivate instrumentation ie. monkey-unpatch. This must run
                # regardless of the response. Keep 'return' clauses below.
                for panel in reversed(toolbar.enabled_panels):
                    panel.disable_instrumentation()

            for panel in reversed(toolbar.enabled_panels):
                panel.generate_stats(request, response)
                panel.generate_server_timing(request, response)

            if getattr(request, 'input_logged_request', False):

                DebugToolbarData.objects.create(
                    logged_request=request.input_logged_request, toolbar=toolbar.render_toolbar()
                )
            return response
        else:
            return self.get_response(request)



import json

from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import TextField
from django.db.models.functions import Cast
from django.template.defaultfilters import truncatechars
from django.utils.html import format_html, format_html_join, mark_safe, format_html_join
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.models import ContentType

from pyston.filters.default_filters import SimpleMethodEqualFilter
from pyston.utils.decorators import filter_by, order_by, filter_class

from is_core.generic_views.inlines.inline_table_views import InlineTableView
from is_core.generic_views.mixins import TabItem, TabsViewMixin
from is_core.generic_views.table_views import TableView
from is_core.main import UIRESTModelISCore
from is_core.utils import render_model_objects_with_link, render_model_object_with_link
from is_core.utils.decorators import short_description

from security.config import settings
from security.models import CommandLog, InputLoggedRequest, OutputLoggedRequest, CeleryTaskLog, CeleryTaskRunLog

from ansi2html import Ansi2HTMLConverter


def display_json(value):
    return json.dumps(value, indent=4, ensure_ascii=False, cls=DjangoJSONEncoder)


def display_as_code(value):
    return format_html('<code style="white-space:pre-wrap;">{}</code>', value) if value else value


def display_related_objects(request, related_objects):
    related_object_instances = []
    for related_object in related_objects:
        try:
            related_object_instances.append(related_object.object)
        except (ObjectDoesNotExist, AttributeError):
            pass

    return render_model_objects_with_link(request, related_object_instances)


def get_content_type_pks_of_parent_related_classes():
    return {
        ContentType.objects.get_for_model(model_class).pk
        for model_class in (CommandLog, InputLoggedRequest, OutputLoggedRequest, CeleryTaskLog, CeleryTaskRunLog)
    }


class UsernameUserFilter(SimpleMethodEqualFilter):

    def get_filter_term(self, value, operator_slug, request):
        user_model = get_user_model()
        return {
            'user_id__in': list(
                user_model.objects.filter(
                   **{'{}__contains'.format(user_model.USERNAME_FIELD): value}
                ).annotate(
                    str_id=Cast('id', output_field=TextField())
                ).values_list('str_id', flat=True)
            )
        }


class SecurityISCoreMixin:

    @short_description(_('related objects'))
    def display_related_objects(self, obj, request):
        return display_related_objects(
            request, obj.related_objects.exclude(object_ct_id__in=get_content_type_pks_of_parent_related_classes())
        )

    @short_description(_('source'))
    def display_source(self, obj, request):
        return display_related_objects(
            request, obj.related_objects.filter(object_ct_id__in=get_content_type_pks_of_parent_related_classes())
        )

    @short_description(_('raised output logged requests'))
    def display_output_logged_requests(self, obj, request):
        return render_model_objects_with_link(
            request,
            OutputLoggedRequest.objects.filter(
                related_objects__object_id=obj.pk,
                related_objects__object_ct_id=ContentType.objects.get_for_model(obj).pk
            )
        )

    @short_description(_('raised command logs'))
    def display_command_logs(self, obj, request):
        return render_model_objects_with_link(
            request,
            CommandLog.objects.filter(
                related_objects__object_id=obj.pk,
                related_objects__object_ct_id=ContentType.objects.get_for_model(obj).pk
            )
        )

    @short_description(_('raised celery task logs'))
    def display_celery_task_logs(self, obj, request):
        return render_model_objects_with_link(
            request,
            CeleryTaskLog.objects.filter(
                related_objects__object_id=obj.pk,
                related_objects__object_ct_id=ContentType.objects.get_for_model(obj).pk
            )
        )


class RequestsLogISCore(SecurityISCoreMixin, UIRESTModelISCore):

    abstract = True

    can_create = can_update = can_delete = False

    @short_description(_('queries'))
    def queries_code(self, obj):
        return display_as_code(display_json(obj.queries)) if obj else None

    @short_description(_('request body'))
    def request_body_code(self, obj):
        return display_as_code(obj.request_body) if obj else None

    @short_description(_('request headers'))
    def request_headers_code(self, obj):
        return display_as_code(display_json(obj.request_headers)) if obj else None

    @short_description(_('response body'))
    def response_body_code(self, obj):
        return display_as_code(obj.response_body) if obj else None

    @short_description(_('response headers'))
    def response_headers_code(self, obj):
        return display_as_code(display_json(obj.response_headers)) if obj else None

    @short_description(_('error description'))
    def error_description_code(self, obj):
        return display_as_code(obj.error_description) if obj else None


class InputRequestsLogISCore(RequestsLogISCore):

    model = InputLoggedRequest
    abstract = True

    ui_list_fields = (
        'id', 'created_at', 'changed_at', 'request_timestamp', 'response_timestamp', 'response_time', 'status',
        'response_code', 'host', 'short_path', 'slug', 'ip', 'user', 'method', 'type', 'short_response_body',
        'short_request_body', 'short_queries', 'short_request_headers'
    )

    form_fieldsets = (
        (_('Request'), {'fields': ('created_at', 'changed_at', 'request_timestamp', 'host', 'method', 'path',
                                   'queries_code', 'request_headers_code', 'request_body_code', 'is_secure')}),
        (_('Response'), {'fields': ('response_timestamp', 'response_code', 'status', 'response_headers_code',
                                    'response_body_code', 'type', 'error_description_code')}),
        (_('User information'), {'fields': ('user', 'ip')}),
        (_('Extra information'), {'fields': ('slug', 'response_time', 'display_related_objects',
                                             'display_output_logged_requests', 'display_command_logs',
                                             'display_celery_task_logs')}),
    )

    def get_form_fieldsets(self, request, obj=None):
        form_fieldsets = list(super().get_form_fieldsets(request, obj))

        app_names = {app.name for app in apps.get_app_configs()}

        if (settings.SHOW_DEBUG_TOOLBAR and 'security.contrib.debug_toolbar_log' in app_names
                and obj and hasattr(obj, 'input_logged_request_toolbar')):
            form_fieldsets.append((None, {'fields': ('debug_toolbar',)}))
        return form_fieldsets

    @short_description(_('user'))
    @filter_class(UsernameUserFilter)
    def user(self, obj):
        return obj.user

    @short_description('')
    def debug_toolbar(self, obj):
        return mark_safe(obj.input_logged_request_toolbar.toolbar)


class OutputRequestsLogISCore(RequestsLogISCore):

    model = OutputLoggedRequest
    abstract = True

    ui_list_fields = (
        'id', 'created_at', 'changed_at', 'request_timestamp', 'response_timestamp', 'response_time', 'status',
        'response_code', 'host', 'short_path', 'method', 'slug', 'short_response_body', 'short_request_body',
        'short_queries', 'short_request_headers'
    )

    form_fieldsets = (
        (_('Request'), {'fields': ('created_at', 'changed_at', 'request_timestamp', 'host', 'method', 'path',
                                   'queries_code', 'request_headers_code', 'request_body_code', 'is_secure')}),
        (_('Response'), {'fields': ('response_timestamp', 'response_code', 'status', 'response_headers_code',
                                    'response_body_code', 'error_description_code')}),
        (_('Extra information'), {'fields': ('slug', 'response_time', 'display_related_objects', 'display_source')}),
    )


class CommandLogISCore(SecurityISCoreMixin, UIRESTModelISCore):

    model = CommandLog

    can_create = can_update = can_delete = False

    ui_list_fields = (
        'id', 'created_at', 'changed_at', 'name', 'start', 'stop', 'time', 'executed_from_command_line', 'is_successful'
    )

    form_fieldsets = (
        (None, {
            'fields': ('created_at', 'changed_at', 'name', 'input', 'output_html', 'error_message',
                       'display_related_objects', 'display_source', 'display_output_logged_requests',
                       'display_command_logs', 'display_celery_task_logs'),
            'class': 'col-sm-6'
        }),
        (None, {
            'fields': ('start', 'stop', 'time', 'executed_from_command_line', 'is_successful'),
            'class': 'col-sm-6'
        }),
    )

    abstract = True

    @short_description(_('output'))
    def output_html(self, obj=None):
        if obj and obj.output is not None:
            conv = Ansi2HTMLConverter()
            output = mark_safe(conv.convert(obj.output, full=False))
            return display_as_code(output)
        return None


class CeleryTaskLogTabs(TabsViewMixin):

    tabs = (
        TabItem('list-celerytasklog', _('celery task')),
        TabItem('list-celerytaskrunlog', _('celery task run')),
    )


class CeleryTaskLogTableView(CeleryTaskLogTabs, TableView):
    pass


class CeleryTaskRunLogISCore(SecurityISCoreMixin, UIRESTModelISCore):

    model = CeleryTaskRunLog

    abstract = True

    can_create = can_update = can_delete = False

    rest_extra_filter_fields = (
        'celery_task_id',
    )

    ui_list_fields = (
        'celery_task_id', 'created_at', 'changed_at', 'name', 'state', 'start', 'stop', 'time', 'result', 'retries',
        'get_task_log'
    )

    form_fields = (
        'celery_task_id', 'task_log', 'start', 'stop', 'time', 'state', 'result', 'error_message', 'output_html',
        'retries', 'estimated_time_of_next_retry', 'display_related_objects', 'display_output_logged_requests',
        'display_command_logs', 'display_celery_task_logs'
    )

    ui_list_view = CeleryTaskLogTableView

    default_ordering = ('-created_at',)

    @short_description(_('celery task log'))
    def task_log(self, obj):
        return obj.get_task_log()

    @short_description(_('output'))
    def output_html(self, obj):
        if obj and obj.output is not None:
            conv = Ansi2HTMLConverter()
            output = mark_safe(conv.convert(obj.output, full=False))
            return display_as_code(output)
        return None


class CeleryTaskRunLogInlineTableView(InlineTableView):

    model = CeleryTaskRunLog
    fields = (
        'created_at', 'changed_at', 'start', 'stop', 'time', 'state', 'result', 'retries'
    )

    def _get_list_filter(self):
        return {
            'filter': {
                'celery_task_id': self.parent_instance.celery_task_id
            }
        }


class CeleryTaskLogISCore(SecurityISCoreMixin, UIRESTModelISCore):

    model = CeleryTaskLog

    abstract = True

    can_create = can_update = can_delete = False

    ui_list_fields = (
        'celery_task_id', 'created_at', 'changed_at', 'name', 'short_input', 'state', 'get_start', 'get_stop',
        'queue_name'
    )

    form_fieldsets = (
        (None, {
            'fields': (
                'celery_task_id', 'created_at', 'changed_at', 'name', 'state', 'get_start', 'get_stop',
                'estimated_time_of_first_arrival', 'expires', 'stale', 'queue_name', 'input', 'display_related_objects',
                'display_source'
            )
        }),
        (_('celery task runs'), {'inline_view': CeleryTaskRunLogInlineTableView}),
    )

    ui_list_view = CeleryTaskLogTableView

    @filter_by('input')
    @order_by('input')
    @short_description(_('input'))
    def short_input(self, obj):
        return truncatechars(obj.input, 50)

    def is_active_menu_item(self, request, active_group):
        return active_group in {
            self.menu_group,
            'celerytaskrunlog',
        }



import atexit
import logging
import re
import sys
import traceback
from contextlib import ContextDecorator
from importlib import import_module
from io import StringIO
from threading import local

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.core.signals import request_finished
from django.db.transaction import get_connection
from django.urls import resolve
from django.urls.exceptions import Resolver404
from django.utils.timezone import now

from .config import settings


output_logged_request_logger = logging.getLogger('security.output_request')


class LogManagementError(Exception):
    pass


class OutputLoggedRequestContext:
    """
    Data structure that stores data for creation OutputLoggedRequest model object
    """

    def __init__(self, data, related_objects=None):
        self.data = data
        self.related_objects = related_objects

    def create(self, logs=None):
        from security.models import OutputLoggedRequest

        logs = [] if logs is None else logs
        output_logged_request = OutputLoggedRequest.objects.create(
            **self.data
        )
        related_objects = list(self.related_objects) + logs
        output_logged_request.related_objects.add(*related_objects)
        return output_logged_request


class LogContextStackFrame:

    def __init__(self, input_logged_request, command_log, celery_task_log, celery_task_run_log,
                 output_requests_related_objects,
                 output_requests_slug):
        self.input_logged_request = input_logged_request
        self.command_log = command_log
        self.celery_task_log = celery_task_log
        self.celery_task_run_log = celery_task_run_log
        self.output_requests_related_objects = (
            list(output_requests_related_objects) if output_requests_related_objects else []
        )
        self.output_requests_slug = output_requests_slug
        self.output_logged_requests = []

    def fork(self, input_logged_request, command_log, celery_task_log, celery_task_run_log,
             output_requests_related_objects, output_requests_slug):
        output_requests_related_objects = (
            list(output_requests_related_objects) if output_requests_related_objects else []
        )
        return LogContextStackFrame(
            input_logged_request or self.input_logged_request,
            command_log or self.command_log,
            celery_task_log or self.celery_task_log,
            celery_task_run_log or self.celery_task_run_log,
            output_requests_related_objects=(
                self.output_requests_related_objects + output_requests_related_objects
            ),
            output_requests_slug=self.output_requests_slug if output_requests_slug is None else output_requests_slug
        )

    def join(self, other_context):
        self.output_logged_requests += other_context.output_logged_requests

    def get_logs(self):
        return list(
            filter(None, [
                self.input_logged_request,
                self.command_log,
                self.celery_task_log,
                self.celery_task_run_log,
            ])
        )

    def save(self):
        for output_logged_request in self.output_logged_requests:
            output_logged_request.create(self.get_logs())


class LogContextManager(local):

    def __init__(self):
        self.clear()
        request_finished.connect(self._request_finished_receiver)
        if 'reversion' in django_settings.INSTALLED_APPS:
            from reversion.signals import post_revision_commit

            post_revision_commit.connect(self._post_revision_commit)

    def is_active(self):
        """Returns whether there is an active revision for this thread."""
        return bool(self._stack)

    def _assert_active(self):
        """Checks for an active revision, throwning an exception if none."""
        if not self.is_active():  # pragma: no cover
            raise LogManagementError('There is no active context log for this thread')

    @property
    def input_logged_request(self):
        return self._current_frame.input_logged_request

    @property
    def command_log(self):
        return self._current_frame.command_log

    @property
    def celery_task_log(self):
        return self._current_frame.celery_task_log

    @property
    def celery_task_run_log(self):
        return self._current_frame.celery_task_run_log

    def get_logs(self):
        return self._current_frame.get_logs()

    @property
    def _current_frame(self):
        self._assert_active()
        return self._stack[-1]

    def start(self, input_logged_request=None, command_log=None, celery_task_log=None, celery_task_run_log=None,
              output_requests_related_objects=None, output_requests_slug=None):
        """
        Begins a context log for this thread.

        This MUST be balanced by a call to `end`
        """
        if self.is_active():
            self._stack.append(
                self._current_frame.fork(
                    input_logged_request,
                    command_log,
                    celery_task_log,
                    celery_task_run_log,
                    output_requests_related_objects,
                    output_requests_slug
                )
            )
        else:
            self._stack.append(
                LogContextStackFrame(
                    input_logged_request,
                    command_log,
                    celery_task_log,
                    celery_task_run_log,
                    output_requests_related_objects or [],
                    output_requests_slug
                )
            )

    def end(self):
        self._assert_active()
        connection = get_connection(settings.LOG_DB_NAME)
        stack_frame = self._stack.pop()
        if self._stack and connection.in_atomic_block:
            self._current_frame.join(stack_frame)
        elif self._stack:
            stack_frame.save()
        else:
            try:
                stack_frame.save()
            finally:
                self.clear()

    def clear(self):
        self._stack = []

    def _request_finished_receiver(self, **kwargs):
        """
        Called at the end of a request, ensuring that any open logs
        are closed. Not closing all active revisions can cause memory leaks
        and weird behaviour.
        """
        while self.is_active():  # pragma: no cover
            self.end()

    def log_output_requests(self, output_logged_request_context):
        connection = get_connection(settings.LOG_DB_NAME)

        if connection.in_atomic_block:
            self._current_frame.output_logged_requests.append(output_logged_request_context)
        else:
            output_logged_request_context.create(logs=self.get_logs())

    def get_output_request_related_objects(self):
        if not self.is_active():
            return []

        return self._current_frame.output_requests_related_objects

    def get_output_request_slug(self):
        if not self.is_active():
            return None

        return self._current_frame.output_requests_slug

    def _post_revision_commit(self, **kwargs):
        """
        Called as a post save of revision model of the reversion library.
        If log context manager is active input logged request, command
        log or celery task run log is joined with revision via related objects.
        """
        revision = kwargs['revision']
        if self.is_active():
            if self.input_logged_request:
                self.input_logged_request.related_objects.add(revision)
            if self.command_log:
                self.command_log.related_objects.add(revision)
            if self.celery_task_run_log:
                self.celery_task_run_log.related_objects.add(revision)


log_context_manager = LogContextManager()


class AtomicLog(ContextDecorator):
    """
    Context decorator that stores logged requests to database connections and inside exit method
    stores it to the database
    """

    def __init__(self, input_logged_request=None, command_log=None, celery_task_log=None, celery_task_run_log=None,
                 output_requests_related_objects=None, output_requests_slug=None):
        self._command_log = command_log
        self._celery_task_log = celery_task_log
        self._celery_task_run_log = celery_task_run_log
        self._input_logged_request = input_logged_request
        self._output_requests_related_objects = (
            [] if output_requests_related_objects is None else list(output_requests_related_objects)
        )
        self._output_requests_slug = output_requests_slug

    def __enter__(self):
        log_context_manager.start(
            self._input_logged_request,
            self._command_log,
            self._celery_task_log,
            self._celery_task_run_log,
            self._output_requests_related_objects,
            self._output_requests_slug,
        )

    def __exit__(self, exc_type, exc_value, traceback):
        log_context_manager.end()


def log_output_request(data, related_objects=None):
    """
    Helper for logging output requests
    :param data: dict of input attributes of OutputLoggedRequest model
    :param related_objects: objects that will be related to OutputLoggedRequest object
    """
    output_logged_request_context = OutputLoggedRequestContext(data, related_objects)
    if log_context_manager.is_active():
        log_context_manager.log_output_requests(output_logged_request_context)
    else:
        output_logged_request_context.create()

    if settings.LOG_OUTPUT_REQUESTS:
        output_logged_request_logger.info(
            ('"%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s"'),
            data.get('request_timestamp', ''),
            data.get('response_timestamp', ''),
            data.get('response_time', ''),
            data.get('response_code', ''),
            data.get('host', ''),
            data.get('path', ''),
            data.get('method', ''),
            data.get('slug', ''),
        )


def is_base_collection(v):
    return isinstance(v, (list, tuple, set))


def get_throttling_validators(name):
    try:
        return getattr(import_module(settings.DEFAULT_THROTTLING_VALIDATORS_PATH), name)
    except (ImportError, AttributeError):
        raise ImproperlyConfigured('Throttling validator configuration {} is not defined'.format(name))


def get_headers(request):
    regex = re.compile('^HTTP_')
    return dict((regex.sub('', header), value) for (header, value)
                in request.META.items() if header.startswith('HTTP_'))


def remove_nul_from_string(value):
    return value.replace('\x00', '')


class LogStringIO(StringIO):

    def __init__(self, post_write_callback=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_newline = 0
        self._post_write_callback = post_write_callback

    def _post_write(self):
        if self._post_write_callback:
            self._post_write_callback(self)

    def write(self, s):
        line_first = True
        for line in s.split('\n'):
            if not line_first:
                super().write('\n')
                self._last_newline = self.tell()
            cursor_first = True
            for cursor_block in line.split('\r'):
                if not cursor_first:
                    self.seek(self._last_newline)
                    self.truncate()
                cursor_first = False
                super().write(cursor_block)
            line_first = False
        self._post_write()

    def isatty(self):
        return True


class TeeStringIO:
    """
    StringIO that additionally writes to another stream(s).
    """
    ANSI_ESCAPE_REGEX = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def __init__(self, out, log_stream):
        self._out = out
        self._log_stream = log_stream
        self._last_newline = 0
        super().__init__()

    def write(self, s):
        if self.isatty():
            self._out.write(s)
        else:
            self._out.write(self.ANSI_ESCAPE_REGEX.sub('', s))
        self._log_stream.write(s)

    def __getattr__(self, name):
        return getattr(self._out, name)

    def isatty(self):
        return hasattr(self._out, 'isatty') and self._out.isatty()

    def flush(self):
        self._out.flush()
        self._log_stream.flush()


class CommandLogger(StringIO):
    """
    A helper class that runs a django command and logs details about its run into DB.
    """

    def __init__(self, command_function, command_args=None, command_kwargs=None, stdout=None, stderr=None, **kwargs):
        """
        Initializes the command logger.

        Arguments:
            command_function: Callable that implements the command.
            stdout: Custom stream where command's standard output will be written.
            stderr: Custom stream where command's error output will be written.
            **kwargs: Keyword arguments passed to CommandLog model.
        """
        assert 'name' in kwargs, 'Key name must be provided in kwargs'

        self.command_function = command_function
        self.command_args = command_args or ()
        self.command_kwargs = command_kwargs or {}

        self.kwargs = kwargs
        self.stdout = stdout if stdout else sys.stdout
        self.stderr = stderr if stderr else sys.stderr

    def _store_log_output(self, output_stream):
        self.command_log.change_and_save(
            output=output_stream.getvalue(),
            update_only_changed_fields=True
        )

    def run(self):
        """
        Runs the command function and returns its return value or re-raises any exceptions. The run of the command will
        not be logged if it is in excluded commands setting.
        """
        from security.models import CommandLog
        from security.decorators import atomic_log

        if self.kwargs['name'] in settings.COMMAND_LOG_EXCLUDED_COMMANDS:
            return self.command_function(
                stdout=self.stdout, stderr=self.stderr, *self.command_args, **self.command_kwargs
            )

        self.command_log = CommandLog.objects.create(start=now(), **self.kwargs)
        if log_context_manager.is_active():
            related_objects = log_context_manager.get_logs()
            for related_object in related_objects:
                self.command_log.related_objects.add(related_object)

        self.output = LogStringIO(post_write_callback=self._store_log_output)
        stdout = TeeStringIO(self.stdout, self.output)
        stderr = TeeStringIO(self.stderr, self.output)

        # register call of the finish method in case the command exits the interpreter prematurely
        atexit.register(lambda: self._finish(error_message='Command was killed'))

        try:
            with atomic_log(command_log=self.command_log):
                ret_val = self.command_function(
                    stdout=stdout, stderr=stderr, *self.command_args, **self.command_kwargs
                )
                self._finish(success=True)
                return ret_val
        except Exception as ex:  # pylint: disable=W0703
            self._finish(success=False, error_message=traceback.format_exc())
            raise ex

    def _finish(self, success=False, error_message=None):
        if self.command_log.stop is None:
            stop = now()
            self.command_log.change_and_save(
                output=self.output.getvalue(),
                stop=stop,
                time=(stop - self.command_log.start).total_seconds(),
                is_successful=success,
                error_message=error_message
            )


def regex_sub_groups_global(pattern, repl, string):
    """
    Globally replace all groups inside pattern with `repl`.
    If `pattern` doesn't have groups the whole match is replaced.
    """
    for search in reversed(list(re.finditer(pattern, string))):
        for i in range(len(search.groups()), 0 if search.groups() else -1, -1):
            start, end = search.span(i)
            string = string[:start] + repl + string[end:]
    return string


def get_view_from_request_or_none(request):
    try:
        return resolve(request.path).func
    except Resolver404:
        return None


is_running_migration = False



# Generated by Django 3.1.4 on 2020-12-21 18:49

from django.db import migrations, models
import django.core.serializers.json


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0001_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='celerytasklog',
            name='task_args',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='task args'),
        ),
        migrations.AlterField(
            model_name='celerytasklog',
            name='task_kwargs',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='task kwargs'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunlog',
            name='result',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='result'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunlog',
            name='task_args',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='task args'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunlog',
            name='task_kwargs',
            field=models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                   null=True, verbose_name='task kwargs'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='queries',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='queries'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='request_headers',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='request headers'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='response_headers',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='response headers'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='queries',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='queries'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='request_headers',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='request headers'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='response_headers',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                   verbose_name='response headers'),
        ),
    ]
